#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Anke Boersma <demm@kaosx.us>
#   SPDX-FileCopyrightText: 2015 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2016 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2018 AlmAck <gluca86@gmail.com>
#   SPDX-FileCopyrightText: 2018-2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import os
import re
import shutil


import libcalamares

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configuring locales.")


RE_IS_COMMENT = re.compile("^ *#")
def is_comment(line):
    """
    Does the @p line look like a comment? Whitespace, followed by a #
    is a comment-only line.
    """
    return bool(RE_IS_COMMENT.match(line))

RE_TRAILING_COMMENT = re.compile("#.*$")
RE_REST_OF_LINE = re.compile("\\s.*$")
def extract_locale(line):
    """
    Extracts a locale from the @p line, and returns a pair of
    (extracted-locale, uncommented line). The locale is the
    first word of the line after uncommenting (in the human-
    readable text explanation at the top of most /etc/locale.gen
    files, the locales may be bogus -- either "" or e.g. "Configuration")
    """
    # Remove leading spaces and comment signs
    line = RE_IS_COMMENT.sub("", line)
    uncommented = line.strip()
    fields = RE_TRAILING_COMMENT.sub("", uncommented).strip().split()
    if len(fields) != 2:
        # Not exactly two fields, can't be a proper locale line
        return "", uncommented
    else:
        # Drop all but first field
        locale = RE_REST_OF_LINE.sub("", uncommented)
        return locale, uncommented


def rewrite_locale_gen(srcfilename, destfilename, locale_conf):
    """
    Copies a locale.gen file from @p srcfilename to @p destfilename
    (this may be the same name), enabling those locales that can
    be found in the map @p locale_conf. Also always enables en_US.UTF-8.
    """
    en_us_locale = 'en_US.UTF-8'

    # Get entire source-file contents
    text = []
    try:
        with open(srcfilename, "r") as gen:
            text = gen.readlines()
    except FileNotFoundError:
        # That's ok, the file doesn't exist so assume empty
        pass

    # we want unique values, so locale_values should have 1 or 2 items
    locale_values = set(locale_conf.values())
    locale_values.add(en_us_locale)  # Always enable en_US as well

    enabled_locales = {}
    seen_locales = set()

    # Write source out again, enabling some
    with open(destfilename, "w") as gen:
        for line in text:
            c = is_comment(line)
            locale, uncommented = extract_locale(line)

            # Non-comment lines are preserved, and comment lines
            # may be enabled if they match a desired locale
            if not c:
                seen_locales.add(locale)
            else:
                for locale_value in locale_values:
                    if locale.startswith(locale_value):
                        enabled_locales[locale] = uncommented
            gen.write(line)

        gen.write("\n###\n#\n# Locales enabled by Calamares\n")
        for locale, line in enabled_locales.items():
            if locale not in seen_locales:
                gen.write(line + "\n")
                seen_locales.add(locale)

        for locale in locale_values:
            if locale not in seen_locales:
                gen.write("# Missing: %s\n" % locale)


def run():
    """ Create locale """
    import libcalamares

    locale_conf = libcalamares.globalstorage.value("localeConf")

    if not locale_conf:
        locale_conf = {
            'LANG': 'en_US.UTF-8',
            'LC_NUMERIC': 'en_US.UTF-8',
            'LC_TIME': 'en_US.UTF-8',
            'LC_MONETARY': 'en_US.UTF-8',
            'LC_PAPER': 'en_US.UTF-8',
            'LC_NAME': 'en_US.UTF-8',
            'LC_ADDRESS': 'en_US.UTF-8',
            'LC_TELEPHONE': 'en_US.UTF-8',
            'LC_MEASUREMENT': 'en_US.UTF-8',
            'LC_IDENTIFICATION': 'en_US.UTF-8'
        }

    install_path = libcalamares.globalstorage.value("rootMountPoint")

    if install_path is None:
        libcalamares.utils.warning("rootMountPoint is empty, {!s}".format(install_path))
        return (_("Configuration Error"),
                _("No root mount point is given for <pre>{!s}</pre> to use." ).format("localecfg"))

    target_locale_gen = "{!s}/etc/locale.gen".format(install_path)
    target_locale_gen_bak = target_locale_gen + ".bak"
    target_locale_conf_path = "{!s}/etc/locale.conf".format(install_path)
    target_etc_default_path = "{!s}/etc/default".format(install_path)

    # restore backup if available
    if os.path.exists(target_locale_gen_bak):
        shutil.copy2(target_locale_gen_bak, target_locale_gen)
        libcalamares.utils.debug("Restored backup {!s} -> {!s}"
            .format(target_locale_gen_bak, target_locale_gen))

    # run locale-gen if detected; this *will* cause an exception
    # if the live system has locale.gen, but the target does not:
    # in that case, fix your installation filesystem.
    if os.path.exists('/etc/locale.gen'):
        rewrite_locale_gen(target_locale_gen, target_locale_gen, locale_conf)
        libcalamares.utils.target_env_call(['locale-gen'])
        libcalamares.utils.debug('{!s} done'.format(target_locale_gen))

    # write /etc/locale.conf
    with open(target_locale_conf_path, "w") as lcf:
        for k, v in locale_conf.items():
            lcf.write("{!s}={!s}\n".format(k, v))
        libcalamares.utils.debug('{!s} done'.format(target_locale_conf_path))

    # write /etc/default/locale if /etc/default exists and is a dir
    if os.path.isdir(target_etc_default_path):
        with open(os.path.join(target_etc_default_path, "locale"), "w") as edl:
            for k, v in locale_conf.items():
                edl.write("{!s}={!s}\n".format(k, v))
        libcalamares.utils.debug('{!s} done'.format(target_etc_default_path))

    return None



#! /usr/bin/env python3
#
#  === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: BSD-2-Clause
#
"""
Python3 script to scrape some data out of ICU CLDR supplemental data.

To use this script, you must have downloaded the CLDR data, e.g.
http://unicode.org/Public/cldr/35.1/, and extracted the zip file.
Run the script from **inside** the common/ durectory that is created
(or fix the hard-coded path).

The script tries to print C++ code that compiles; if there are encoding
problems, it will print some kind of representation of the problematic
lines.

To avoid having to cross-reference multiple XML files, the script
cheats: it reads the comments as well to get names. So it looks for
pairs of lines like this:

   <likelySubtag from="und_BQ" to="pap_Latn_BQ"/>
   <!--{ ?; ?; Caribbean Netherlands } => { Papiamento; Latin; Caribbean Netherlands }-->

It extracts the 2-character country code "BQ" from the sub-tag, and
parses the comment to get a language and country name (instead of looking up
"pap" and "BQ" in other tables). This may be considered a hack.

A large collection of exceptions can be found in the two *_mapper tables,
which massage the CLDR names to Qt enum values.
"""
#
### END USAGE

import sys

# These are languages listed in CLDR that don't match
# the enum-values in QLocale::Language.
language_mapper = {
    "?" : "AnyLanguage",
    "Bangla" : "Bengali",
    "Kalaallisut" : "Greenlandic",
    "Haitian Creole" : "Haitian",
    "Kyrgyz" : "Kirghiz",
    "Norwegian Bokmål" : "NorwegianBokmal",
    "Tokelau" : "TokelauLanguage",
    "Tuvalu" : "TuvaluLanguage",
    }

country_mapper = {
    "Åland Islands" : "AlandIslands",
    "St. Barthélemy" : "SaintBarthelemy",
    "Côte d’Ivoire" : "IvoryCoast",
    "Curaçao" : "CuraSao",
    "Réunion" : "Reunion",
    "São Tomé & Príncipe" : "SaoTomeAndPrincipe",
    "Bosnia & Herzegovina" : "BosniaAndHerzegowina",
    "Czechia" : "CzechRepublic",
    "St. Pierre & Miquelon" : "SaintPierreAndMiquelon",
    "Vatican City" : "VaticanCityState",
    "South Georgia & South Sandwich Islands" : "SouthGeorgiaAndTheSouthSandwichIslands",
    "Timor-Leste" : "EastTimor",
    "Wallis & Futuna" : "WallisAndFutunaIslands",
    "Myanmar (Burma)" : "Myanmar",
    "Svalbard & Jan Mayen" : "SvalbardAndJanMayenIslands",
    "St. Martin" : "SaintMartin",
    "North Macedonia" : "Macedonia",
    "Hong Kong SAR China" : "HongKong",
    "Macao SAR China" : "Macau",
    "Eurozone" : "AnyCountry",  # Not likely for GeoIP
    "Caribbean Netherlands" : "Bonaire",  # Bonaire, Saba, St.Eustatius
    }

class CountryData:
    def __init__(self, country_code, language_name, country_name):
        """
        Takes a 2-letter country name, and enum names from
        QLocale::Language and QLocale::Country. An empty
        @p country code is acceptable, for the terminating
        entry in the data array (and yields a 0,0 code).
        """
        if country_code:
            assert len(country_code) == 2
            self.country_code = country_code
            self.language_enum = language_name
            self.country_enum = country_name
        else:
            self.country_code = ""
            self.language_enum = "AnyLanguage"
            self.country_enum = "AnyCountry"

    def __str__(self):
        if self.country_code:
            char0 = "'{!s}'".format(self.country_code[0])
            char1 = "'{!s}'".format(self.country_code[1])
        else:
            char0 = "0"
            char1 = "0"

        return "{!s} QLocale::Language::{!s}, QLocale::Country::{!s}, {!s}, {!s} {!s},".format(
            "{",
            self.language_enum,
            self.country_enum,
            char0,
            char1,
            "}")

    # Must match type name below
    cpp_classname = "CountryData"

    # Must match the output format of __str__ above
    cpp_declaration = """
struct CountryData
{
    QLocale::Language l;
    QLocale::Country c;
    char cc1;
    char cc2;
};
"""


def extricate_subtags(l1, l2):
    """
    Given two lines @p l1 and @p l2 which are the <likelySubtag> element-line
    and the comment-line underneath it, return a CountryData for them,
    or None if the two lines are not relevant (e.g. not the right subtag from,
    or 3-letter country codes.
    """
    if 'from="und_' not in l1:
        return
    if '{ ?; ?;' not in l2:
        return

    # This is extremely crude "parsing" which chops up the string
    # by delimiter and then extracts some substring.
    l1_parts = l1.split("und_")
    l2_parts = l2.split(";")

    l1_first_quote = l1_parts[1].find('"')
    l1_code = l1_parts[1][:l1_first_quote]
    if len(l1_code) != 2:
        return

    l2_brace = l2_parts[2].find("{")
    l2_language = l2_parts[2][l2_brace+1:].strip()
    l2_brace = l2_parts[2].find("}")
    l2_country = l2_parts[2][:l2_brace-1].strip()

    # Handle mapped cases
    l2_language = language_mapper.get(l2_language, l2_language)
    l2_language = l2_language.replace(" ", "")

    # Handle mapped cases and then do a bunch of standard replacements.
    l2_country = country_mapper.get(l2_country, l2_country)
    l2_country = l2_country.replace(" ", "").replace("-", "").replace(".","").replace("&","And")

    return CountryData(l1_code, l2_language, l2_country)


def read_subtags_file():
    """
    Returns a list of CountryData objects from the likelySubtags file.
    """
    data = []

    with open("supplemental/likelySubtags.xml", "rt", encoding="UTF-8") as f:
        l1 = "a line"
        while l1:
            l1 = f.readline()
            if '<likelySubtag from="und_' not in l1:
                continue
            l2 = f.readline()

            if l1:
                assert "likelySubtag" in l1, l1;
                assert "<!--" in l2, l2;

                data.append(extricate_subtags(l1, l2))

    data.append(CountryData("", None, None))
    return [c for c in data if c is not None]


cpp_header_comment = """/*   GENERATED FILE DO NOT EDIT
*
*  === This file is part of Calamares - <https://calamares.io> ===
*
* SPDX-FileCopyrightText: 1991-2019 Unicode, Inc.
* SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
* SPDX-License-Identifier: CC0
*
* This file is derived from CLDR data from Unicode, Inc. Applicable terms
* are listed at http://unicode.org/copyright.html , of which the most
* important are:
*
* A. Unicode Copyright
*    1. Copyright © 1991-2019 Unicode, Inc. All rights reserved.
* B. Definitions
*    Unicode Data Files ("DATA FILES") include all data files under the directories:
*    https://www.unicode.org/Public/
* C. Terms of Use
*    1. Certain documents and files on this website contain a legend indicating
*       that "Modification is permitted." Any person is hereby authorized,
*       without fee, to modify such documents and files to create derivative
*       works conforming to the Unicode® Standard, subject to Terms and
*       Conditions herein.
*    2. Any person is hereby authorized, without fee, to view, use, reproduce,
*       and distribute all documents and files, subject to the Terms and
*       Conditions herein.
*/

// BEGIN Generated from CLDR data
// *INDENT-OFF*
// clang-format off

"""

cpp_footer_comment = """
// END Generated from CLDR data
"""


def make_identifier(classname):
    """
    Given a class name (e.g. CountryData) return an identifer
    for the data-table for that class.
    """
    identifier = [ classname[0].lower() ]
    for c in classname[1:]:
        if c.isupper():
            identifier.extend(["_", c.lower()])
        else:
            identifier.append(c)

    return "".join(identifier)


def export_class(cls, data):
    """
    Given a @p cls and a list of @p data objects from that class,
    print (to stdout) a C++ file for that data.
    """
    identifier = make_identifier(cls.cpp_classname)

    with open("{!s}_p.cpp".format(cls.cpp_classname), "wt", encoding="UTF-8") as f:
        f.write(cpp_header_comment)
        f.write(cls.cpp_declaration)
        f.write("\nstatic constexpr int const {!s}_size = {!s};\n".format(
            identifier,
            len(data)))
        f.write("\nstatic const {!s} {!s}_table[] = {!s}\n".format(
            cls.cpp_classname,
            identifier,
            "{"))
        for d in data:
            f.write(str(d))
            f.write("\n")
        f.write("};\n\n");
        f.write("static_assert( (sizeof({!s}_table) / sizeof({!s})) == {!s}_size, \"Table size mismatch for {!s}\" );\n\n".format(
            identifier,
            cls.cpp_classname,
            identifier,
            cls.cpp_classname))
        f.write(cpp_footer_comment)


if __name__ == "__main__":
    export_class(CountryData, read_subtags_file())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2017 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

"""
=== Example Python jobmodule.

A Python jobmodule is a Python program which imports libcalamares and
has a function run() as entry point. run() must return None if everything
went well, or a tuple (str,str) with an error message and description
if something went wrong.
"""

import libcalamares
import os
from time import gmtime, strftime, sleep

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Dummy python job.")

status = _("Dummy python step {}").format(0)

def pretty_status_message():
    return status

def run():
    """Dummy python job."""
    libcalamares.utils.debug("LocaleDir=" +
                             str(libcalamares.utils.gettext_path()))
    libcalamares.utils.debug("Languages=" +
                             str(libcalamares.utils.gettext_languages()))

    os.system("/bin/sh -c \"touch ~/calamares-dummypython\"")
    accumulator = strftime("%Y-%m-%d %H:%M:%S", gmtime()) + "\n"
    accumulator += "Calamares version: " + libcalamares.VERSION_SHORT + "\n"
    accumulator += "This job's name: " + libcalamares.job.pretty_name + "\n"
    accumulator += "This job's path: " + libcalamares.job.working_path
    libcalamares.utils.debug(accumulator)

    accumulator = "*** Job configuration "
    accumulator += str(libcalamares.job.configuration)
    libcalamares.utils.debug(accumulator)

    accumulator = "*** globalstorage test ***"
    accumulator += "lala: "
    accumulator += str(libcalamares.globalstorage.contains("lala")) + "\n"
    accumulator += "foo: "
    accumulator += str(libcalamares.globalstorage.contains("foo")) + "\n"
    accumulator += "count: " + str(libcalamares.globalstorage.count())
    libcalamares.utils.debug(accumulator)

    libcalamares.globalstorage.insert("item2", "value2")
    libcalamares.globalstorage.insert("item3", 3)
    accumulator = "keys: {}\n".format(str(libcalamares.globalstorage.keys()))
    libcalamares.utils.debug(accumulator)

    accumulator = "remove: {}\n".format(
       str(libcalamares.globalstorage.remove("item2")))
    accumulator += "values: {} {} {}\n".format(
        str(libcalamares.globalstorage.value("foo")),
        str(libcalamares.globalstorage.value("item2")),
        str(libcalamares.globalstorage.value("item3")))
    libcalamares.utils.debug(accumulator)

    libcalamares.utils.debug("Run dummy python")

    sleep(1)

    try:
        configlist = list(libcalamares.job.configuration["a_list"])
    except KeyError:
        configlist = ["no list"]

    global status
    c = 1
    for k in configlist:
        status = _("Dummy python step {}").format(str(c) + ":" + repr(k))
        libcalamares.utils.debug(_("Dummy python step {}").format(str(k)))
        sleep(1)
        libcalamares.job.setprogress(c * 1.0 / (len(configlist)+1))
        c += 1

    sleep(3)

    # To indicate an error, return a tuple of:
    # (message, detailed-error-message)
    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2019 Collabora Ltd <arnaud.ferraris@collabora.com>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares
import os
import stat
import subprocess
from time import gmtime, strftime, sleep
from math import gcd

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext

def pretty_name():
    return _("Installing data.")

def lcm(a, b):
    """
    Computes the Least Common Multiple of 2 numbers
    """
    return a * b / gcd(a, b)

def get_device_size(device):
    """
    Returns a filesystem's total size and block size in bytes.
    For block devices, block size is the device's block size.
    For other files (fs images), block size is 1 byte.

    @param device: str
        Absolute path to the device or filesystem image.
    @return: tuple(int, int)
        The filesystem's size and its block size.
    """
    mode = os.stat(device).st_mode
    if stat.S_ISBLK(mode):
        basedevice = ""
        partition = os.path.basename(device)
        tmp = partition
        while len(tmp) > 0:
            tmp = tmp[:-1]
            if os.path.exists("/sys/block/" + tmp):
                basedevice = tmp
                break
        # Get device block size
        file = open("/sys/block/" + basedevice + "/queue/hw_sector_size")
        blocksize = int(file.readline())
        file.close()
        # Get partition size
        file = open("/sys/block/" + basedevice + "/" + partition + "/size")
        size = int(file.readline()) * blocksize
        file.close()
    else:
        size = os.path.getsize(device)
        blocksize = 1

    return size, blocksize

class RawFSLowSpaceError(Exception):
    pass

class RawFSItem:
    __slots__ = ['source', 'destination', 'filesystem', 'resize']

    def copy(self, current=0, total=1):
        """
        Copies a raw filesystem on a disk partition, and grow it to the full destination
        partition's size if required.

        @param current: int
            The index of the current item in the filesystems list
            (used for progress reporting)
        @param total: int
            The number of items in the filesystems list
            (used for progress reporting)
        """
        count = 0

        libcalamares.utils.debug("Copying {} to {}".format(self.source, self.destination))
        if libcalamares.job.configuration.get("bogus", False):
            return

        srcsize, srcblksize = get_device_size(self.source)
        destsize, destblksize = get_device_size(self.destination)

        if destsize < srcsize:
            raise RawFSLowSpaceError
            return

        # Compute transfer block size (100x the LCM of the block sizes seems a good fit)
        blksize = int(100 * lcm(srcblksize, destblksize))

        # Execute copy
        src = open(self.source, "rb")
        dest = open(self.destination, "wb")
        buffer = src.read(blksize)
        while len(buffer) > 0:
            dest.write(buffer)
            count += len(buffer)
            # Compute job progress
            progress = ((count / srcsize) + (current)) / total
            libcalamares.job.setprogress(progress)
            # Read next data block
            buffer = src.read(blksize)
        src.close()
        dest.close()

        if self.resize:
            if "ext" in self.filesystem:
                libcalamares.utils.debug("Resizing filesystem on {}".format(self.destination))
                subprocess.run(["e2fsck", "-f", "-y", self.destination])
                subprocess.run(["resize2fs", self.destination])

    def __init__(self, config, device, fs):
        libcalamares.utils.debug("Adding an entry for raw copy of {} to {}".format(
                config["source"], device))
        self.source = os.path.realpath(config["source"])
        # If source is a mount point, look for the actual device mounted on it
        if os.path.ismount(self.source):
            procmounts = open("/proc/mounts", "r")
            for line in procmounts:
                if self.source in line.split():
                    self.source = line.split()[0]
                    break

        self.destination = device
        self.filesystem = fs
        try:
            self.resize = bool(config["resize"])
        except KeyError:
            self.resize = False

def update_global_storage(item, gs):
    for partition in gs:
        if partition["device"] == item.destination:
            ret = subprocess.run(["blkid", "-s", "UUID", "-o", "value", item.destination],
                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
            if ret.returncode == 0:
                libcalamares.utils.debug("Setting {} UUID to {}".format(item.destination,
                        ret.stdout.rstrip()))
                gs[gs.index(partition)]["uuid"] = ret.stdout.rstrip()
                gs[gs.index(partition)]["source"] = item.source

    libcalamares.globalstorage.remove("partitions")
    libcalamares.globalstorage.insert("partitions", gs)

def run():
    """Raw filesystem copy module"""
    filesystems = list()
    partitions = libcalamares.globalstorage.value("partitions")

    if not partitions:
        libcalamares.utils.warning("partitions is empty, {!s}".format(partitions))
        return (_("Configuration Error"),
                _("No partitions are defined for <pre>{!s}</pre> to use." ).format("rawfs"))

    libcalamares.utils.debug("Copying {!s} raw partitions.".format(len(partitions)))
    for partition in partitions:
        if partition["mountPoint"]:
            for src in libcalamares.job.configuration["targets"]:
                if src["mountPoint"] == partition["mountPoint"]:
                    filesystems.append(RawFSItem(src, partition["device"], partition["fs"]))

    for item in filesystems:
        try:
            item.copy(filesystems.index(item), len(filesystems))
        except RawFSLowSpaceError:
            return ("Not enough free space",
                "{} partition is too small to copy {} on it".format(item.destination, item.source))
        update_global_storage(item, partitions)

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2016-2017 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import platform

from PythonQt.QtGui import *
import PythonQt.calamares as calamares

# WARNING: the Calamares PythonQt API is considered EXPERIMENTAL as of
# Calamares 2.5. It comes with no promise or commitment to API stability.


# Set up translations.
# You may skip this if your Calamares module has no user visible strings.
# DO NOT install _ into the builtin namespace because each module loads
# its own catalog.
# DO use the gettext class-based API and manually alias _ as described in:
# https://docs.python.org/3.5/library/gettext.html#localizing-your-module
import gettext
import inspect
import os
_filename = inspect.getframeinfo(inspect.currentframe()).filename
_path = os.path.dirname(os.path.abspath(_filename))

_ = gettext.gettext

# Example Python ViewModule.
# A Python ViewModule is a Python program which defines a ViewStep class.
# One UI module ==> one ViewStep.
# This class must be marked with the @calamares_module decorator. A
# ViewModule may define other classes, but only one may be decorated with
# @calamares_module. Such a class must conform to the Calamares ViewStep
# interface and functions as the entry point of the module.
# A ViewStep manages one or more "wizard pages" through methods like
# back/next, and reports its status through isNextEnabled/isBackEnabled/
# isAtBeginning/isAtEnd. The whole UI, including all the pages, must be
# exposed as a single QWidget, returned by the widget function.
#
# For convenience, both C++ and PythonQt ViewSteps are considered to be
# implementations of ViewStep.h. Additionally, the Calamares PythonQt API
# allows Python developers to keep their identifiers more Pythonic on the
# Python side. Thus, all of the following are considered valid method
# identifiers in a ViewStep implementation: isNextEnabled, isnextenabled,
# is_next_enabled.


@calamares_module
class DummyPythonQtViewStep:
    def __init__(self):
        # Importing PythonQt.QtGui provides access to most Qt widget classes.
        self.main_widget = QFrame()

        self.main_widget.setLayout(QVBoxLayout())

        label = QLabel()
        self.main_widget.layout().addWidget(label)

        accumulator = "\nCalamares+PythonQt running embedded Python " +\
                      platform.python_version()
        label.text = accumulator

        btn = QPushButton()

        # Python strings can be used wherever a method wants a QString. Python
        # gettext translations can be used seamlessly as well.
        btn.setText(_("Click me!"))
        self.main_widget.layout().addWidget(btn)

        # The syntax for signal-slot connections is very simple, though
        # slightly different from the C++ equivalent. There are no SIGNAL and
        # SLOT macros, and a signal can be connected to any Python method
        # (without a special "slot" designation).
        btn.connect("clicked(bool)", self.on_btn_clicked)

    def on_btn_clicked(self):
        self.main_widget.layout().addWidget(QLabel(_("A new QLabel.")))

    def prettyName(self):
        return _("Dummy PythonQt ViewStep")

    def isNextEnabled(self):
        return True  # The "Next" button should be clickable

    def isBackEnabled(self):
        return True  # The "Back" button should be clickable

    def isAtBeginning(self):
        # True means the currently shown UI page is the first page of this
        # module, thus a "Back" button click will not be handled by this
        # module and will cause a skip to the previous ViewStep instead
        # (if any). False means that the present ViewStep provides other UI
        # pages placed logically "before" the current one, thus a "Back" button
        # click will be handled by this module instead of skipping to another
        # ViewStep. A module (ViewStep) with only one page will always return
        # True here.
        return True

    def isAtEnd(self):
        # True means the currently shown UI page is the last page of this
        # module, thus a "Next" button click will not be handled by this
        # module and will cause a skip to the next ViewStep instead (if any).
        # False means that the present ViewStep provides other UI pages placed
        # logically "after" the current one, thus a "Next" button click will
        # be handled by this module instead of skipping to another ViewStep.
        # A module (ViewStep) with only one page will always return True here.
        return True

    def jobs(self):
        # Returns a list of objects that implement Calamares::Job.
        return [DummyPQJob("Dummy PythonQt job reporting for duty")]

    def widget(self):
        # Returns the base QWidget of this module's UI.
        return self.main_widget

    def retranslate(self, locale_name):
        # This is where it gets slightly weird. In most desktop applications we
        # shouldn't need this kind of mechanism, because we could assume that
        # the operating environment is configured to use a certain language.
        # Usually the user would change the system-wide language in a settings
        # UI, restart the application, done.
        # Alas, Calamares runs on an unconfigured live system, and one of the
        # core features of Calamares is to allow the user to pick a language.
        # Unfortunately, strings in the UI do not automatically react to a
        # runtime language change. To get UI strings in a new language, all
        # user-visible strings must be retranslated (by calling tr() in C++ or
        # _() in Python) and reapplied on the relevant widgets.
        # When the user picks a new UI translation language, Qt raises a QEvent
        # of type LanguageChange, which propagates through the QObject
        # hierarchy. By catching and reacting to this event, we can show
        # user-visible strings in the new language at the right time.
        # The C++ side of the Calamares PythonQt API catches the LanguageChange
        # event and calls the present method. It is then up to the module
        # developer to add here all the needed code to load the module's
        # translation catalog for the new language (which is separate from the
        # main Calamares strings catalog) and reapply any user-visible strings.
        calamares.utils.debug("PythonQt retranslation event "
                              "for locale name: {}".format(locale_name))

        # First we load the catalog file for the new language...
        try:
            global _
            _t = gettext.translation('dummypythonqt',
                                     localedir=os.path.join(_path, 'lang'),
                                     languages=[locale_name])
            _ = _t.gettext
        except OSError as e:
            calamares.utils.debug(e)
            pass

        # ... and then we can call setText(_("foo")) and similar methods on
        # the relevant widgets here to reapply the strings.

# An example Job class. Implements Calamares::Job. For method identifiers, the
# same rules apply as for ViewStep. No decorators are necessary here, because
# only the ViewStep implementation is the unique entry point, and a module can
# have any number of jobs.


class DummyPQJob:
    def __init__(self, my_msg):
        self.my_msg = my_msg

    def pretty_name(self):
        return _("The Dummy PythonQt Job")

    def pretty_description(self):
        return _("This is the Dummy PythonQt Job. "
                 "The dummy job says: {}").format(self.my_msg)

    def pretty_status_message(self):
        return _("A status message for Dummy PythonQt Job.")

    def exec(self):
        # As an example, we touch a file in the target root filesystem.
        rmp = calamares.global_storage['rootMountPoint']
        os.system("touch {}/calamares_dpqt_was_here".format(rmp))
        calamares.utils.debug("the dummy job says {}".format(self.my_msg))
        return {'ok': True}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014-2015 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2015-2017 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2017 2019, Adriaan de Groot <groot@kde.org>
#   SPDX-FileCopyrightText: 2017-2018 Gabriel Craciunescu <crazy@frugalware.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares
import os
import re

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configure GRUB.")


def get_grub_config_path(root_mount_point):
    """
    Figures out where to put the grub config files. Returns
    a the full path of a file inside that
    directory, as "the config file".

    Returns a path into @p root_mount_point.
    """
    default_dir = os.path.join(root_mount_point, "etc/default")
    default_config_file = "grub"

    if "prefer_grub_d" in libcalamares.job.configuration and libcalamares.job.configuration["prefer_grub_d"]:
        possible_dir = os.path.join(root_mount_point, "etc/default/grub.d")
        if os.path.exists(possible_dir) and os.path.isdir(possible_dir):
            default_dir = possible_dir
            default_config_file = "00calamares"

    if not os.path.exists(default_dir):
        try:
            os.mkdir(default_dir)
        except:
            libcalamares.utils.debug("Failed to create '%r'" % default_dir)
            raise

    return os.path.join(default_dir, default_config_file)


def get_zfs_root():
    """
    Looks in global storage to find the zfs root

    :return: A string containing the path to the zfs root or None if it is not found
    """

    zfs = libcalamares.globalstorage.value("zfsDatasets")

    if not zfs:
        libcalamares.utils.warning("Failed to locate zfs dataset list")
        return None

    # Find the root dataset
    for dataset in zfs:
        try:
            if dataset["mountpoint"] == "/":
                return dataset["zpool"] + "/" + dataset["dsName"]
        except KeyError:
            # This should be impossible
            libcalamares.utils.warning("Internal error handling zfs dataset")
            raise

    return None


def modify_grub_default(partitions, root_mount_point, distributor):
    """
    Configures '/etc/default/grub' for hibernation and plymouth.

    @see bootloader/main.py, for similar handling of kernel parameters

    :param partitions:
    :param root_mount_point:
    :param distributor: name of the distributor to fill in for
        GRUB_DISTRIBUTOR. Must be a string. If the job setting
        *keepDistributor* is set, then this is only used if no
        GRUB_DISTRIBUTOR is found at all (otherwise, when *keepDistributor*
        is set, the GRUB_DISTRIBUTOR lines are left unchanged).
        If *keepDistributor* is unset or false, then GRUB_DISTRIBUTOR
        is always updated to set this value.
    :return:
    """
    default_grub = get_grub_config_path(root_mount_point)
    distributor_replace = distributor.replace("'", "'\\''")
    dracut_bin = libcalamares.utils.target_env_call(
        ["sh", "-c", "which dracut"]
        )
    plymouth_bin = libcalamares.utils.target_env_call(
        ["sh", "-c", "which plymouth"]
        )

    # Shell exit value 0 means success
    have_plymouth = plymouth_bin == 0
    have_dracut = dracut_bin == 0

    use_splash = ""
    swap_uuid = ""
    swap_outer_uuid = ""
    swap_outer_mappername = None
    no_save_default = False
    unencrypted_separate_boot = any(p["mountPoint"] == "/boot" and "luksMapperName" not in p for p in partitions)
    # If there is no dracut, and the root partition is ZFS, this gets set below
    zfs_root_path = None

    for partition in partitions:
        if partition["mountPoint"] in ("/", "/boot") and partition["fs"] in ("btrfs", "f2fs"):
            no_save_default = True
            break

    if have_plymouth:
        use_splash = "splash"

    cryptdevice_params = []

    if have_dracut:
        for partition in partitions:
            if partition["fs"] == "linuxswap" and not partition.get("claimed", None):
                # Skip foreign swap
                continue
            has_luks = "luksMapperName" in partition
            if partition["fs"] == "linuxswap" and not has_luks:
                swap_uuid = partition["uuid"]

            if (partition["fs"] == "linuxswap" and has_luks):
                swap_outer_uuid = partition["luksUuid"]
                swap_outer_mappername = partition["luksMapperName"]

            if (partition["mountPoint"] == "/" and has_luks):
                cryptdevice_params = [
                    "rd.luks.uuid={!s}".format(partition["luksUuid"])
                    ]
    else:
        for partition in partitions:
            if partition["fs"] == "linuxswap" and not partition.get("claimed", None):
                # Skip foreign swap
                continue
            has_luks = "luksMapperName" in partition
            if partition["fs"] == "linuxswap" and not has_luks:
                swap_uuid = partition["uuid"]

            if (partition["fs"] == "linuxswap" and has_luks):
                swap_outer_mappername = partition["luksMapperName"]

            if (partition["mountPoint"] == "/" and has_luks):
                cryptdevice_params = [
                    "cryptdevice=UUID={!s}:{!s}".format(
                        partition["luksUuid"], partition["luksMapperName"]
                        ),
                    "root=/dev/mapper/{!s}".format(
                        partition["luksMapperName"]
                        )
                ]

            if partition["fs"] == "zfs" and partition["mountPoint"] == "/":
                zfs_root_path = get_zfs_root()

    kernel_params = libcalamares.job.configuration.get("kernel_params", ["quiet"])

    # Currently, grub doesn't detect this properly so it must be set manually
    if zfs_root_path:
        kernel_params.insert(0, "zfs=" + zfs_root_path)

    if cryptdevice_params:
        kernel_params.extend(cryptdevice_params)

    if use_splash:
        kernel_params.append(use_splash)

    if swap_uuid:
        kernel_params.append("resume=UUID={!s}".format(swap_uuid))

    if have_dracut and swap_outer_uuid:
        kernel_params.append("rd.luks.uuid={!s}".format(swap_outer_uuid))
    if swap_outer_mappername:
        kernel_params.append("resume=/dev/mapper/{!s}".format(
            swap_outer_mappername))

    distributor_line = "GRUB_DISTRIBUTOR='{!s}'".format(distributor_replace)

    have_kernel_cmd = False
    have_distributor_line = False

    if "overwrite" in libcalamares.job.configuration:
        overwrite = libcalamares.job.configuration["overwrite"]
    else:
        overwrite = False

    if os.path.exists(default_grub) and not overwrite:
        with open(default_grub, 'r') as grub_file:
            lines = [x.strip() for x in grub_file.readlines()]

        for i in range(len(lines)):
            if lines[i].startswith("#GRUB_CMDLINE_LINUX_DEFAULT"):
                kernel_cmd = "GRUB_CMDLINE_LINUX_DEFAULT=\"{!s}\"".format(
                    " ".join(kernel_params)
                    )
                lines[i] = kernel_cmd
                have_kernel_cmd = True
            elif lines[i].startswith("GRUB_CMDLINE_LINUX_DEFAULT"):
                regex = re.compile(r"^GRUB_CMDLINE_LINUX_DEFAULT\s*=\s*")
                line = regex.sub("", lines[i])
                line = line.lstrip()
                line = line.lstrip("\"")
                line = line.lstrip("'")
                line = line.rstrip()
                line = line.rstrip("\"")
                line = line.rstrip("'")
                existing_params = line.split()

                for existing_param in existing_params:
                    existing_param_name = existing_param.split("=")[0]

                    # the only ones we ever add
                    if existing_param_name not in [
                            "quiet", "resume", "splash"]:
                        kernel_params.append(existing_param)

                kernel_cmd = "GRUB_CMDLINE_LINUX_DEFAULT=\"{!s}\"".format(
                    " ".join(kernel_params)
                    )
                lines[i] = kernel_cmd
                have_kernel_cmd = True
            elif (lines[i].startswith("#GRUB_DISTRIBUTOR")
                  or lines[i].startswith("GRUB_DISTRIBUTOR")):
                if libcalamares.job.configuration.get("keepDistributor", False):
                    lines[i] = distributor_line
                    have_distributor_line = True
                else:
                    # We're not updating because of *keepDistributor*, but if
                    # this was a comment line, then it's still not been set.
                    have_distributor_line = have_distributor_line or not lines[i].startswith("#")
            # If btrfs or f2fs is used, don't save default
            if no_save_default and lines[i].startswith("GRUB_SAVEDEFAULT="):
                lines[i] = "#GRUB_SAVEDEFAULT=\"true\""
    else:
        lines = []

        if "defaults" in libcalamares.job.configuration:
            for key, value in libcalamares.job.configuration[
                    "defaults"].items():
                if value.__class__.__name__ == "bool":
                    if value:
                        escaped_value = "true"
                    else:
                        escaped_value = "false"
                else:
                    escaped_value = str(value).replace("'", "'\\''")

                lines.append("{!s}='{!s}'".format(key, escaped_value))

    if not have_kernel_cmd:
        kernel_cmd = "GRUB_CMDLINE_LINUX_DEFAULT=\"{!s}\"".format(
            " ".join(kernel_params)
            )
        lines.append(kernel_cmd)

    if not have_distributor_line:
        lines.append(distributor_line)

    if cryptdevice_params and not unencrypted_separate_boot:
        lines.append("GRUB_ENABLE_CRYPTODISK=y")

    with open(default_grub, 'w') as grub_file:
        grub_file.write("\n".join(lines) + "\n")

    return None


def run():
    """
    Calls routine with given parameters to modify '/etc/default/grub'.

    :return:
    """

    fw_type = libcalamares.globalstorage.value("firmwareType")

    if (libcalamares.globalstorage.value("bootLoader") is None
            and fw_type != "efi"):
        return None

    partitions = libcalamares.globalstorage.value("partitions")

    if fw_type == "efi":
        esp_found = False

        for partition in partitions:
            if (partition["mountPoint"]
                    == libcalamares.globalstorage.value("efiSystemPartition")):
                esp_found = True

        if not esp_found:
            return None

    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")
    branding = libcalamares.globalstorage.value("branding")
    distributor = branding["bootloaderEntryName"]

    return modify_grub_default(partitions, root_mount_point, distributor)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#   SPDX-FileCopyrightText: 2016 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import os
import re

import libcalamares

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Writing fstab.")


FSTAB_HEADER = """# /etc/fstab: static file system information.
#
# Use 'blkid' to print the universally unique identifier for a device; this may
# be used with UUID= as a more robust way to name devices that works even if
# disks are added and removed. See fstab(5).
#
# <file system>             <mount point>  <type>  <options>  <dump>  <pass>"""

CRYPTTAB_HEADER = """# /etc/crypttab: mappings for encrypted partitions.
#
# Each mapped device will be created in /dev/mapper, so your /etc/fstab
# should use the /dev/mapper/<name> paths for encrypted devices.
#
# See crypttab(5) for the supported syntax.
#
# NOTE: You need not list your root (/) partition here, but it must be set up
#       beforehand by the initramfs (/etc/mkinitcpio.conf). The same applies
#       to encrypted swap, which should be set up with mkinitcpio-openswap
#       for resume support.
#
# <name>               <device>                         <password> <options>"""

# Turn Parted filesystem names into fstab names
FS_MAP = {
    "fat16": "vfat",
    "fat32": "vfat",
    "linuxswap": "swap",
}


def mkdir_p(path):
    """ Create directory.

    :param path:
    """
    if not os.path.exists(path):
        os.makedirs(path)


def is_ssd_disk(disk_name):
    """ Checks if given disk is actually a ssd disk.

    :param disk_name:
    :return:
    """
    filename = os.path.join("/sys/block", disk_name, "queue/rotational")

    if not os.path.exists(filename):
        # Should not happen unless sysfs changes, but better safe than sorry
        return False

    with open(filename) as sysfile:
        return sysfile.read() == "0\n"


def disk_name_for_partition(partition):
    """ Returns disk name for each found partition.

    :param partition:
    :return:
    """
    name = os.path.basename(partition["device"])

    if name.startswith("mmcblk") or name.startswith("nvme"):
        # Typical mmc device is mmcblk0p1, nvme looks like nvme0n1p2
        return re.sub("p[0-9]+$", "", name)

    return re.sub("[0-9]+$", "", name)


class FstabGenerator(object):
    """ Class header

    :param partitions:
    :param root_mount_point:
    :param mount_options:
    :param ssd_extra_mount_options:
    :param crypttab_options:
    :param tmp_options:
    """
    def __init__(self, partitions, root_mount_point, mount_options,
                 ssd_extra_mount_options, crypttab_options, tmp_options):
        self.partitions = partitions
        self.root_mount_point = root_mount_point
        self.mount_options = mount_options
        self.ssd_extra_mount_options = ssd_extra_mount_options
        self.crypttab_options = crypttab_options
        self.tmp_options = tmp_options
        self.ssd_disks = set()
        self.root_is_ssd = False

    def run(self):
        """ Calls needed sub routines.

        :return:
        """
        self.find_ssd_disks()
        self.generate_fstab()
        self.generate_crypttab()
        self.create_mount_points()

        return None

    def find_ssd_disks(self):
        """ Checks for ssd disks """
        disks = {disk_name_for_partition(x) for x in self.partitions}
        self.ssd_disks = {x for x in disks if is_ssd_disk(x)}

    def generate_crypttab(self):
        """ Create crypttab. """
        mkdir_p(os.path.join(self.root_mount_point, "etc"))
        crypttab_path = os.path.join(self.root_mount_point, "etc", "crypttab")

        with open(crypttab_path, "w") as crypttab_file:
            print(CRYPTTAB_HEADER, file=crypttab_file)

            for partition in self.partitions:
                dct = self.generate_crypttab_line_info(partition)

                if dct:
                    self.print_crypttab_line(dct, file=crypttab_file)

    def generate_crypttab_line_info(self, partition):
        """ Generates information for each crypttab entry. """
        if "luksMapperName" not in partition or "luksUuid" not in partition:
            return None

        mapper_name = partition["luksMapperName"]
        luks_uuid = partition["luksUuid"]
        if not mapper_name or not luks_uuid:
            return None

        password = "/crypto_keyfile.bin"
        crypttab_options = self.crypttab_options

        # Set crypttab password for partition to none and remove crypttab options
        # if root partition was not encrypted
        if any([p["mountPoint"] == "/"
                and "luksMapperName" not in p
                for p in self.partitions]):
            password = "none"
            crypttab_options = ""
        # on root partition when /boot is unencrypted
        elif partition["mountPoint"] == "/":
            if any([p["mountPoint"] == "/boot"
                   and "luksMapperName" not in p
                   for p in self.partitions]):
                password = "none"
                crypttab_options = ""

        return dict(
            name=mapper_name,
            device="UUID=" + luks_uuid,
            password=password,
            options=crypttab_options,
        )

    def print_crypttab_line(self, dct, file=None):
        """ Prints line to '/etc/crypttab' file. """
        line = "{:21} {:<45} {} {}".format(dct["name"],
                                           dct["device"],
                                           dct["password"],
                                           dct["options"],
                                           )

        print(line, file=file)

    def generate_fstab(self):
        """ Create fstab. """
        mkdir_p(os.path.join(self.root_mount_point, "etc"))
        fstab_path = os.path.join(self.root_mount_point, "etc", "fstab")

        with open(fstab_path, "w") as fstab_file:
            print(FSTAB_HEADER, file=fstab_file)

            for partition in self.partitions:
                # Special treatment for a btrfs subvolumes
                if (partition["fs"] == "btrfs"
                   and partition["mountPoint"] == "/"):
                    # Subvolume list has been created in mount.conf and curated in mount module,
                    # so all subvolumes here should be safe to add to fstab
                    btrfs_subvolumes = libcalamares.globalstorage.value("btrfsSubvolumes")
                    for s in btrfs_subvolumes:
                        mount_entry = partition
                        mount_entry["mountPoint"] = s["mountPoint"]
                        mount_entry["subvol"] = s["subvolume"]
                        dct = self.generate_fstab_line_info(mount_entry)
                        if dct:
                            self.print_fstab_line(dct, file=fstab_file)
                elif partition["fs"] != "zfs":  # zfs partitions don't need an entry in fstab
                    dct = self.generate_fstab_line_info(partition)
                    if dct:
                        self.print_fstab_line(dct, file=fstab_file)

            if self.root_is_ssd:
                # Old behavior was to mount /tmp as tmpfs
                # New behavior is to use tmpOptions to decide
                # if mounting /tmp as tmpfs and which options to use
                ssd = self.tmp_options.get("ssd", {})
                if not ssd:
                    ssd = self.tmp_options.get("default", {})
                # Default to True to mimic old behavior
                tmpfs = ssd.get("tmpfs", True)

                if tmpfs:
                    options = ssd.get("options", "defaults,noatime,mode=1777")
                    # Mount /tmp on a tmpfs
                    dct = dict(device="tmpfs",
                               mount_point="/tmp",
                               fs="tmpfs",
                               options=options,
                               check=0,
                               )
                    self.print_fstab_line(dct, file=fstab_file)

    def generate_fstab_line_info(self, partition):
        """
        Generates information (a dictionary of fstab-fields)
        for the given @p partition.
        """
        # Some "fs" names need special handling in /etc/fstab, so remap them.
        filesystem = partition["fs"].lower()
        filesystem = FS_MAP.get(filesystem, filesystem)
        luks_mapper_name = partition.get("luksMapperName", None)
        mount_point = partition["mountPoint"]
        disk_name = disk_name_for_partition(partition)
        is_ssd = disk_name in self.ssd_disks

        # Swap partitions are called "linuxswap" by parted.
        # That "fs" is visible in GS, but that gets mapped
        # to "swap", above, because that's the spelling needed in /etc/fstab
        if not mount_point and not filesystem == "swap":
            return None
        if not mount_point:
            mount_point = "swap"

        if filesystem == "swap" and not partition.get("claimed", None):
            libcalamares.utils.debug("Ignoring foreign swap {!s} {!s}".format(disk_name, partition.get("uuid", None)))
            return None

        # If this is btrfs subvol a dedicated to a swapfile, use different options than a normal btrfs subvol
        if filesystem == "btrfs" and partition.get("subvol", None) == "/@swap":
            options = self.get_mount_options("btrfs_swap", mount_point)
        else:
            options = self.get_mount_options(filesystem, mount_point)

        if is_ssd:
            extra = self.ssd_extra_mount_options.get(filesystem)

            if extra:
                options += "," + extra

        if mount_point == "/" and filesystem != "btrfs":
            check = 1
        elif mount_point and mount_point != "swap" and filesystem != "btrfs":
            check = 2
        else:
            check = 0

        if mount_point == "/":
            self.root_is_ssd = is_ssd

        # If there's a set-and-not-empty subvolume set, add it
        if filesystem == "btrfs" and partition.get("subvol",None):
            options = "subvol={},".format(partition["subvol"]) + options

        device = None
        if luks_mapper_name:
            device = "/dev/mapper/" + luks_mapper_name
        elif partition["uuid"]:
            device = "UUID=" + partition["uuid"]
        else:
            device = partition["device"]

        if not device:
            # TODO: we get here when the user mounted a previously encrypted partition
            # This should be catched early in the process
            return None

        return dict(device=device,
                    mount_point=mount_point,
                    fs=filesystem,
                    options=options,
                    check=check,
                    )

    def print_fstab_line(self, dct, file=None):
        """ Prints line to '/etc/fstab' file. """
        line = "{:41} {:<14} {:<7} {:<10} 0 {}".format(dct["device"],
                                                       dct["mount_point"],
                                                       dct["fs"],
                                                       dct["options"],
                                                       dct["check"],
                                                       )
        print(line, file=file)

    def create_mount_points(self):
        """ Creates mount points """
        for partition in self.partitions:
            if partition["mountPoint"]:
                mkdir_p(self.root_mount_point + partition["mountPoint"])

    def get_mount_options(self, filesystem, mount_point):
        efiMountPoint = libcalamares.globalstorage.value("efiSystemPartition")
        job_config = libcalamares.job.configuration

        if (mount_point == efiMountPoint and "efiMountOptions" in job_config):
            return job_config["efiMountOptions"]

        return self.mount_options.get(filesystem,
                                      self.mount_options["default"])


def create_swapfile(root_mount_point, root_btrfs):
    """
    Creates /swapfile in @p root_mount_point ; if the root filesystem
    is on btrfs, then handle some btrfs specific features as well,
    as documented in
        https://wiki.archlinux.org/index.php/Swap#Swap_file

    The swapfile-creation covers progress from 0.2 to 0.5
    """
    libcalamares.job.setprogress(0.2)
    if root_btrfs:
        # btrfs swapfiles must reside on a subvolume that is not snapshotted to prevent file system corruption
        swapfile_path = os.path.join(root_mount_point, "swap/swapfile")
        with open(swapfile_path, "wb") as f:
            pass
        libcalamares.utils.host_env_process_output(["chattr", "+C", "+m", swapfile_path])  # No Copy-on-Write, no compression
    else:
        swapfile_path = os.path.join(root_mount_point, "swapfile")
        with open(swapfile_path, "wb") as f:
            pass
    # Create the swapfile; swapfiles are small-ish
    zeroes = bytes(16384)
    with open(swapfile_path, "wb") as f:
        total = 0
        desired_size = 512 * 1024 * 1024  # 512MiB
        while total < desired_size:
            chunk = f.write(zeroes)
            if chunk < 1:
                libcalamares.utils.debug("Short write on {!s}, cancelling.".format(swapfile_path))
                break
            libcalamares.job.setprogress(0.2 + 0.3 * ( total / desired_size ) )
            total += chunk
    os.chmod(swapfile_path, 0o600)
    libcalamares.utils.host_env_process_output(["mkswap", swapfile_path])
    libcalamares.job.setprogress(0.5)


def run():
    """ Configures fstab.

    :return:
    """
    global_storage = libcalamares.globalstorage
    conf = libcalamares.job.configuration
    partitions = global_storage.value("partitions")
    root_mount_point = global_storage.value("rootMountPoint")

    if not partitions:
        libcalamares.utils.warning("partitions is empty, {!s}"
                                   .format(partitions))
        return (_("Configuration Error"),
                _("No partitions are defined for <pre>{!s}</pre> to use.")
                .format("fstab"))
    if not root_mount_point:
        libcalamares.utils.warning("rootMountPoint is empty, {!s}"
                                   .format(root_mount_point))
        return (_("Configuration Error"),
                _("No root mount point is given for <pre>{!s}</pre> to use.")
                .format("fstab"))

    # This follows the GS settings from the partition module's Config object
    swap_choice = global_storage.value( "partitionChoices" )
    if swap_choice:
        swap_choice = swap_choice.get( "swap", None )
        if swap_choice and swap_choice == "file":
            # There's no formatted partition for it, so we'll sneak in an entry
            root_partitions = [ p["fs"].lower() for p in partitions if p["mountPoint"] == "/" ]
            root_btrfs = (root_partitions[0] == "btrfs") if root_partitions else False
            if root_btrfs:
                partitions.append( dict(fs="swap", mountPoint=None, claimed=True, device="/swap/swapfile", uuid=None) )
            else:
                partitions.append( dict(fs="swap", mountPoint=None, claimed=True, device="/swapfile", uuid=None) )
        else:
            swap_choice = None

    libcalamares.job.setprogress(0.1)
    mount_options = conf.get("mountOptions", {})
    ssd_extra_mount_options = conf.get("ssdExtraMountOptions", {})
    crypttab_options = conf.get("crypttabOptions", "luks")
    tmp_options = conf.get("tmpOptions", {})

    # We rely on mount_options having a default; if there wasn't one,
    # bail out with a meaningful error.
    if not mount_options:
        libcalamares.utils.warning("No mount options defined, {!s} partitions".format(len(partitions)))
        return (_("Configuration Error"),
                _("No <pre>{!s}</pre> configuration is given for <pre>{!s}</pre> to use.")
                .format("mountOptions", "fstab"))

    generator = FstabGenerator(partitions,
                               root_mount_point,
                               mount_options,
                               ssd_extra_mount_options,
                               crypttab_options,
                               tmp_options)

    if swap_choice is not None:
        libcalamares.job.setprogress(0.2)
        root_partitions = [ p["fs"].lower() for p in partitions if p["mountPoint"] == "/" ]
        root_btrfs = (root_partitions[0] == "btrfs") if root_partitions else False
        create_swapfile(root_mount_point, root_btrfs)

    try:
        libcalamares.job.setprogress(0.5)
        return generator.run()
    finally:
        libcalamares.job.setprogress(1.0)



#   SPDX-FileCopyrightText: no
#   SPDX-License-Identifier: CC0-1.0
#
# Stubs for part of the Python API from libcalamares
# (although the **actual** API is presented through
# Boost::Python, not as a bare C-extension) so that
# pylint doesn't complain about libcalamares internals.

def count(): return 1

def keys(): return []

def contains(_): return True

def value(key):
    if key in ("branding",):
        return dict()
    if key in ("partitions",):
        return list()
    return ""

def insert(key, value): pass

def remove(_): pass



#! /usr/bin/env python3
#
#  === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: BSD-2-Clause
#
"""
Python3 script to scrape some data out of zoneinfo/zone.tab.

To use this script, you must have a zone.tab in a standard location,
/usr/share/zoneinfo/zone.tab (this is usual on FreeBSD and Linux).

Prints out a few tables of zone names for use in translations.
"""

def scrape_file(file, regionset, zoneset):
    for line in file.readlines():
        if line.startswith("#"):
            continue
        parts = line.split("\t")
        if len(parts) < 3:
            continue

        zoneid = parts[2]
        if not "/" in zoneid:
            continue

        region, zone = zoneid.split("/", 1)

        zone = zone.strip().replace("_", " ")

        regionset.add(region)
        assert(zone not in zoneset)
        zoneset.add(zone)

def write_set(file, label, set):
    file.write("/* This returns a reference to local, which is a terrible idea.\n * Good thing it's not meant to be compiled.\n */\n")
    # Note {{ is an escaped { for Python string formatting
    file.write("static const QStringList& {!s}_table()\n{{\n\treturn QStringList {{\n".format(label))
    for x in sorted(set):
        file.write("""\t\tQObject::tr("{!s}", "{!s}"),\n""".format(x, label))
    file.write("\t\tQString()\n\t};\n}\n\n")

cpp_header_comment = """/*   GENERATED FILE DO NOT EDIT
*
*  === This file is part of Calamares - <https://calamares.io> ===
*
* SPDX-FileCopyrightText: 2009 Arthur David Olson
* SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
* SPDX-License-Identifier: CC0-1.0
*
* This file is derived from zone.tab, which has its own copyright statement:
*
* This file is in the public domain, so clarified as of
* 2009-05-17 by Arthur David Olson.
*
* From Paul Eggert (2018-06-27):
* This file is intended as a backward-compatibility aid for older programs.
* New programs should use zone1970.tab.  This file is like zone1970.tab (see
* zone1970.tab's comments), but with the following additional restrictions:
*
* 1.  This file contains only ASCII characters.
* 2.  The first data column contains exactly one country code.
*
*/

/** THIS FILE EXISTS ONLY FOR TRANSLATIONS PURPOSES **/

// *INDENT-OFF*
// clang-format off
"""

if __name__ == "__main__":
    regions=set()
    zones=set()
    with open("/usr/share/zoneinfo/zone.tab", "r") as f:
        scrape_file(f, regions, zones)
    with open("ZoneData_p.cpp", "w") as f:
        f.write(cpp_header_comment)
        write_set(f, "tz_regions", regions)
        write_set(f, "tz_names", zones)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014-2015 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2017-2018 Gabriel Craciunescu <crazy@frugalware.org>
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Setting hardware clock.")


def run():
    """
    Set hardware clock.
    """
    hwclock_rtc = ["hwclock", "--systohc", "--utc"]
    hwclock_isa = ["hwclock", "--systohc", "--utc", "--directisa"]
    is_broken_rtc = False
    is_broken_isa = False

    ret = libcalamares.utils.target_env_call(hwclock_rtc)
    if ret != 0:
        is_broken_rtc = True
        libcalamares.utils.debug("Hwclock returned error code {}".format(ret))
        libcalamares.utils.debug("  .. RTC method failed, trying ISA bus method.")
    else:
        libcalamares.utils.debug("Hwclock set using RTC method.")
    if is_broken_rtc:
        ret = libcalamares.utils.target_env_call(hwclock_isa)
        if  ret != 0:
            is_broken_isa = True
            libcalamares.utils.debug("Hwclock returned error code {}".format(ret))
            libcalamares.utils.debug("  .. ISA bus method failed.")
        else:
            libcalamares.utils.debug("Hwclock set using ISA bus method.")
    if is_broken_rtc and is_broken_isa:
        libcalamares.utils.debug("BIOS or Kernel BUG: Setting hwclock failed.")

    return None



#   SPDX-FileCopyrightText: no
#   SPDX-License-Identifier: CC0-1.0
#
# Calamares Boilerplate
import libcalamares
libcalamares.globalstorage = libcalamares.GlobalStorage(None)
libcalamares.globalstorage.insert("testing", True)

# Module prep-work
from src.modules.displaymanager import main
default_desktop_environment = main.DesktopEnvironment("startplasma-x11", "kde-plasma.desktop")

# Specific DM test
d = main.DMsddm("/tmp")
d.set_autologin("d", True, default_desktop_environment)
# .. and again (this time checks load/save)
d.set_autologin("d", True, default_desktop_environment)
d.set_autologin("d", True, default_desktop_environment)



#! /usr/bin/env python3
#
#  === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2020 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: BSD-2-Clause
#
"""
Python3 script to scrape x keyboard layout file and produce translations.

To use this script, you must have a base.lst in a standard location,
/usr/local/share/X11/xkb/rules/base.lst (this is usual on FreeBSD).

Prints out a few tables of keyboard model, layout, variant names for
use in translations.
"""

def scrape_file(file, modelsset, layoutsset, variantsset):
    import re
    # These RE's match what is in keyboardglobal.cpp
    model_re = re.compile("^\\s+(\\S+)\\s+(\\w.*)\n$")
    layout_re = re.compile("^\\s+(\\S+)\\s+(\\w.*)\n$")
    variant_re = re.compile("^\\s+(\\S+)\\s+(\\S+): (\\w.*)\n$")

    MODEL, LAYOUT, VARIANT = range(3)
    state = None
    for line in file.readlines():
        # Handle changes in section
        if line.startswith("! model"):
            state = MODEL
            continue
        elif line.startswith("! layout"):
            state = LAYOUT
            continue
        elif line.startswith("! variant"):
            state = VARIANT
            continue
        elif not line.strip():
            state = None
        # Unchanged from last blank
        if state is None:
            continue

        if state == MODEL:
            m = model_re.match(line)
            name = m.groups()[1]
            modelsset.add(name)
        if state == LAYOUT:
            l = layout_re.match(line)
            name = l.groups()[1]
            layoutsset.add(name)
        if state == VARIANT:
            v = variant_re.match(line)
            name = v.groups()[2]
            variantsset.add(name)


def write_set(file, label, set):
    file.write("/* This returns a reference to local, which is a terrible idea.\n * Good thing it's not meant to be compiled.\n */\n")
    # Note {{ is an escaped { for Python string formatting
    file.write("class {!s} : public QObject {{\nQ_OBJECT\npublic:\n".format(label))
    file.write("\tconst QStringList& table()\n\t{\n\treturn QStringList {\n")
    for x in sorted(set):
        file.write("""\t\ttr("{!s}", "{!s}"),\n""".format(x, label))
    file.write("\t\tQString()\n\t};\n}\n}\n\n")

cpp_header_comment = """/*   GENERATED FILE DO NOT EDIT
*
*  === This file is part of Calamares - <https://calamares.io> ===
*
* SPDX-FileCopyrightText: no
* SPDX-License-Identifier: CC0-1.0
*
* This file is derived from base.lst in the Xorg distribution
*
*/

/** THIS FILE EXISTS ONLY FOR TRANSLATIONS PURPOSES **/

// *INDENT-OFF*
// clang-format off
"""

if __name__ == "__main__":
    models=set()
    layouts=set()
    variants=set()
    variants.add( "Default" )
    with open("/usr/local/share/X11/xkb/rules/base.lst", "r") as f:
        scrape_file(f, models, layouts, variants)
    with open("KeyboardData_p.cxxtr", "w") as f:
        f.write(cpp_header_comment)
        write_set(f, "kb_models", models)
        write_set(f, "kb_layouts", layouts)
        write_set(f, "kb_variants", variants)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014-2015 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaid@siduction.org>
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares
from libcalamares.utils import target_env_call


import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Creating initramfs with dracut.")


def run_dracut():
    """
    Creates initramfs, even when initramfs already exists.

    :return:
    """
    return target_env_call(['dracut', '-f'])


def run():
    """
    Starts routine to create initramfs. It passes back the exit code
    if it fails.

    :return:
    """
    return_code = run_dracut()

    if return_code != 0:
        return ( _("Failed to run dracut on the target"),
                 _("The exit code was {}").format(return_code) )



# SPDX-FileCopyrightText: no
# SPDX-License-Identifier: CC0-1.0
#
# Calamares Boilerplate
import libcalamares
libcalamares.globalstorage = libcalamares.GlobalStorage(None)
libcalamares.globalstorage.insert("testing", True)

# Module prep-work
from src.modules.bootloader import main

# Specific Bootloader test
g = main.get_efi_suffix_generator("derp@@SERIAL@@")
assert g is not None
assert g.next() == "derp"  # First time, no suffix
for n in range(9):
    print(g.next())
# We called next() 10 times in total, starting from 0
assert g.next() == "derp10"

g = main.get_efi_suffix_generator("derp@@RANDOM@@")
assert g is not None
for n in range(10):
    print(g.next())
# it's random, nothing to assert

g = main.get_efi_suffix_generator("derp@@PHRASE@@")
assert g is not None
for n in range(10):
    print(g.next())
# it's random, nothing to assert

# Check invalid things
try:
    g = main.get_efi_suffix_generator("derp")
    raise TypeError("Shouldn't get generator (no indicator)")
except ValueError as e:
    pass

try:
    g = main.get_efi_suffix_generator("derp@@HEX@@")
    raise TypeError("Shouldn't get generator (unknown indicator)")
except ValueError as e:
    pass

try:
    g = main.get_efi_suffix_generator("derp@@SERIAL@@x")
    raise TypeError("Shouldn't get generator (trailing garbage)")
except ValueError as e:
    pass

try:
    g = main.get_efi_suffix_generator("derp@@SERIAL@@@@RANDOM@@")
    raise TypeError("Shouldn't get generator (multiple indicators)")
except ValueError as e:
    pass


# Try the generator (assuming no calamares- test files exist in /tmp)
import os
assert "calamares-single" == main.change_efi_suffix("/tmp", "calamares-single")
assert "calamares-serial" == main.change_efi_suffix("/tmp", "calamares-serial@@SERIAL@@")
try:
    os.makedirs("/tmp/calamares-serial", exist_ok=True)
    assert "calamares-serial1" == main.change_efi_suffix("/tmp", "calamares-serial@@SERIAL@@")
finally:
    os.rmdir("/tmp/calamares-serial")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014-2018 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2014-2015 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2014 Kevin Kofler <kevin.kofler@chello.at>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2017 Bernhard Landauer <oberon@manjaro.org>
#   SPDX-FileCopyrightText: 2017 2019, Adriaan de Groot <groot@kde.org>
#   SPDX-FileCopyrightText: 2019 Dominic Hayes <ferenosdev@outlook.com>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import abc
import os
import libcalamares

from libcalamares.utils import gettext_path, gettext_languages

import gettext
_translation = gettext.translation("calamares-python",
                                   localedir=gettext_path(),
                                   languages=gettext_languages(),
                                   fallback=True)
_ = _translation.gettext
_n = _translation.ngettext

class DesktopEnvironment:
    """
    Desktop Environment -- some utility functions for a desktop
    environment (e.g. finding out if it is installed). This
    is independent of the *Display Manager*, which is what
    we're configuring in this module.
    """
    def __init__(self, exec, desktop):
        self.executable = exec
        self.desktop_file = desktop

    def _search_executable(self, root_mount_point, pathname):
        """
        Search for @p pathname within @p root_mount_point .
        If the pathname is absolute, just check there inside
        the target, otherwise earch in a sort-of-sensible $PATH.

        Returns the full (including @p root_mount_point) path
        to that executable, or None.
        """
        if pathname.startswith("/"):
            path = [""]
        else:
            path = ["/bin/", "/usr/bin/", "/sbin/", "/usr/local/bin/"]

        for p in path:
            absolute_path = "{!s}{!s}{!s}".format(root_mount_point, p, pathname)
            if os.path.exists(absolute_path):
                return absolute_path
        return None

    def _search_tryexec(self, root_mount_point, absolute_desktop_file):
        """
        Check @p absolute_desktop_file for a TryExec line and, if that is
        found, search for the command (executable pathname) within
        @p root_mount_point. The .desktop file must live within the
        target root.

        Returns the full (including @p root_mount_point) for the executable
        from TryExec, or None.
        """
        assert absolute_desktop_file.startswith(root_mount_point)
        with open(absolute_desktop_file, "r") as f:
            for tryexec_line in [x for x in f.readlines() if x.startswith("TryExec")]:
                try:
                    key, value = tryexec_line.split("=")
                    if key.strip() == "TryExec":
                        return self._search_executable(root_mount_point, value.strip())
                except:
                    pass
        return None

    def find_executable(self, root_mount_point):
        """
        Returns the full path of the configured executable within @p root_mount_point,
        or None if it isn't found. May search in a semi-sensible $PATH.
        """
        return self._search_executable(root_mount_point, self.executable)

    def find_desktop_file(self, root_mount_point):
        """
        Returns the full path of the .desktop file within @p root_mount_point,
        or None if it isn't found.  Searches both X11 and Wayland sessions.
        """
        x11_sessions = "{!s}/usr/share/xsessions/{!s}.desktop".format(root_mount_point, self.desktop_file)
        wayland_sessions = "{!s}/usr/share/wayland-sessions/{!s}.desktop".format(root_mount_point, self.desktop_file)
        for candidate in (x11_sessions, wayland_sessions):
            if os.path.exists(candidate):
                return candidate
        return None

    def is_installed(self, root_mount_point):
        """
        Check if this environment is installed in the
        target system at @p root_mount_point.
        """
        desktop_file = self.find_desktop_file(root_mount_point)
        if desktop_file is None:
            return False

        return (self.find_executable(root_mount_point) is not None or
                self._search_tryexec(root_mount_point, desktop_file) is not None)

    def update_from_desktop_file(self, root_mount_point):
        """
        Find thie DE in the target system at @p root_mount_point.
        This can update the *executable* configuration value if
        the configured executable isn't found but the TryExec line
        from the .desktop file is.

        The .desktop file is mandatory for a DE.

        Returns True if the DE is installed.
        """
        desktop_file = self.find_desktop_file(root_mount_point)
        if desktop_file is None:
            return False

        executable_file = self.find_executable(root_mount_point)
        if executable_file is not None:
            # .desktop found and executable as well.
            return True

        executable_file = self._search_tryexec(root_mount_point, desktop_file)
        if executable_file is not None:
            # Found from the .desktop file, so update own executable config
            if root_mount_point and executable_file.startswith(root_mount_point):
                executable_file = executable_file[len(root_mount_point):]
            if not executable_file:
                # Somehow chopped down to nothing
                return False

            if executable_file[0] != "/":
                executable_file = "/" + executable_file
            self.executable = executable_file
            return True
        # This is to double-check
        return self.is_installed(root_mount_point)


# This is the list of desktop environments that Calamares looks
# for; if no default environment is **explicitly** configured
# in the `displaymanager.conf` then the first one from this list
# that is found, is used.
#
# Each DE has a sample executable to look for, and a .desktop filename.
# If the executable exists, the DE is assumed to be installed
# and to use the given .desktop filename.
#
# If the .desktop file exists and contains a TryExec line and that
# TryExec executable exists (searched in /bin, /usr/bin, /sbin and
# /usr/local/bin) then the DE is assumed to be installed
# and to use that .desktop filename.
desktop_environments = [
    DesktopEnvironment('/usr/bin/startplasma-x11', 'plasma'),  # KDE Plasma 5.17+
    DesktopEnvironment('/usr/bin/startkde', 'plasma'),  # KDE Plasma 5
    DesktopEnvironment('/usr/bin/startkde', 'kde-plasma'),  # KDE Plasma 4
    DesktopEnvironment(
        '/usr/bin/budgie-desktop', 'budgie-desktop'  # Budgie v10
        ),
    DesktopEnvironment(
        '/usr/bin/budgie-session', 'budgie-desktop'  # Budgie v8
        ),
    DesktopEnvironment('/usr/bin/gnome-session', 'gnome'),
    DesktopEnvironment('/usr/bin/cinnamon-session-cinnamon', 'cinnamon'),
    DesktopEnvironment('/usr/bin/mate-session', 'mate'),
    DesktopEnvironment('/usr/bin/enlightenment_start', 'enlightenment'),
    DesktopEnvironment('/usr/bin/lxsession', 'LXDE'),
    DesktopEnvironment('/usr/bin/startlxde', 'LXDE'),
    DesktopEnvironment('/usr/bin/lxqt-session', 'lxqt'),
    DesktopEnvironment('/usr/bin/pekwm', 'pekwm'),
    DesktopEnvironment('/usr/bin/pantheon-session', 'pantheon'),
    DesktopEnvironment('/usr/bin/startdde', 'deepin'),
    DesktopEnvironment('/usr/bin/startxfce4', 'xfce'),
    DesktopEnvironment('/usr/bin/openbox-session', 'openbox'),
    DesktopEnvironment('/usr/bin/i3', 'i3'),
    DesktopEnvironment('/usr/bin/awesome', 'awesome'),
    DesktopEnvironment('/usr/bin/bspwm', 'bspwm'),
    DesktopEnvironment('/usr/bin/herbstluftwm', 'herbstluftwm'),
    DesktopEnvironment('/usr/bin/qtile', 'qtile'),
    DesktopEnvironment('/usr/bin/xmonad', 'xmonad'),
    DesktopEnvironment('/usr/bin/dwm', 'dwm'),
    DesktopEnvironment('/usr/bin/jwm', 'jwm'),
    DesktopEnvironment('/usr/bin/icewm-session', 'icewm-session'),
    DesktopEnvironment('/usr/bin/fvwm3', 'fvwm3'),
    DesktopEnvironment('/usr/bin/sway', 'sway'),
    DesktopEnvironment('/usr/bin/ukui-session', 'ukui'),
    DesktopEnvironment('/usr/bin/cutefish-session', 'cutefish-xsession'),
]


def find_desktop_environment(root_mount_point):
    """
    Checks which desktop environment is currently installed.

    :param root_mount_point:
    :return:
    """
    libcalamares.utils.debug("Using rootMountPoint {!r}".format(root_mount_point))
    for desktop_environment in desktop_environments:
        if desktop_environment.is_installed(root_mount_point):
            libcalamares.utils.debug(".. selected DE {!s}".format(desktop_environment.desktop_file))
            return desktop_environment
    return None


class DisplayManager(metaclass=abc.ABCMeta):
    """
    Display Manager -- a base class for DM configuration.
    """
    name = None
    executable = None

    def __init__(self, root_mount_point):
        self.root_mount_point = root_mount_point

    def have_dm(self):
        """
        Is this DM installed in the target system?
        The default implementation checks for `executable`
        in the target system.
        """
        if self.executable is None:
            return False

        bin_path = "{!s}/usr/bin/{!s}".format(self.root_mount_point, self.executable)
        sbin_path = "{!s}/usr/sbin/{!s}".format(self.root_mount_point, self.executable)
        return os.path.exists(bin_path) or os.path.exists(sbin_path)

    # The four abstract methods below are called in the order listed here.
    # They must all be implemented by subclasses, but not all of them
    # actually do something for all DMs.

    @abc.abstractmethod
    def basic_setup(self):
        """
        Do basic setup (e.g. users, groups, directory creation) for this DM.
        """
        # Some implementations do nothing

    @abc.abstractmethod
    def desktop_environment_setup(self, desktop_environment):
        """
        Configure the given @p desktop_environment as the default one, in
        the configuration files for this DM.
        """
        # Many implementations do nothing

    @abc.abstractmethod
    def greeter_setup(self):
        """
        Additional setup for the greeter.
        """
        # Most implementations do nothing

    @abc.abstractmethod
    def set_autologin(self, username, do_autologin, default_desktop_environment):
        """
        Configure the DM inside the given @p root_mount_point with
        autologin (if @p do_autologin is True) for the given @p username.
        If the DM supports it, set the default DE to @p default_desktop_environment
        as well.
        """


class DMmdm(DisplayManager):
    name = "mdm"
    executable = "mdm"

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        # Systems with MDM as Desktop Manager
        mdm_conf_path = os.path.join(self.root_mount_point, "etc/mdm/custom.conf")

        if os.path.exists(mdm_conf_path):
            with open(mdm_conf_path, 'r') as mdm_conf:
                text = mdm_conf.readlines()

            with open(mdm_conf_path, 'w') as mdm_conf:
                for line in text:
                    if 'AutomaticLogin=' in line:
                        line = ""
                    if 'AutomaticLoginEnable=True' in line:
                        line = ""
                    if '[daemon]' in line:
                        if do_autologin:
                            line = (
                                "[daemon]\n"
                                "AutomaticLogin={!s}\n"
                                "AutomaticLoginEnable=True\n".format(username)
                                )
                        else:
                            line = (
                                "[daemon]\n"
                                "AutomaticLoginEnable=False\n"
                                )

                    mdm_conf.write(line)
        else:
            with open(mdm_conf_path, 'w') as mdm_conf:
                mdm_conf.write(
                    '# Calamares - Configure automatic login for user\n'
                    )
                mdm_conf.write('[daemon]\n')

                if do_autologin:
                    mdm_conf.write("AutomaticLogin={!s}\n".format(username))
                    mdm_conf.write('AutomaticLoginEnable=True\n')
                else:
                    mdm_conf.write('AutomaticLoginEnable=False\n')

    def basic_setup(self):
        if libcalamares.utils.target_env_call(
                    ['getent', 'group', 'mdm']
                    ) != 0:
            libcalamares.utils.target_env_call(
                ['groupadd', '-g', '128', 'mdm']
                )

        if libcalamares.utils.target_env_call(
                ['getent', 'passwd', 'mdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['useradd',
                    '-c', '"Linux Mint Display Manager"',
                    '-u', '128',
                    '-g', 'mdm',
                    '-d', '/var/lib/mdm',
                    '-s', '/usr/bin/nologin',
                    'mdm'
                    ]
                )

        libcalamares.utils.target_env_call(
            ['passwd', '-l', 'mdm']
            )
        libcalamares.utils.target_env_call(
            ['chown', 'root:mdm', '/var/lib/mdm']
            )
        libcalamares.utils.target_env_call(
            ['chmod', '1770', '/var/lib/mdm']
            )

    def desktop_environment_setup(self, default_desktop_environment):
        os.system(
            "sed -i \"s|default.desktop|{!s}.desktop|g\" "
            "{!s}/etc/mdm/custom.conf".format(
                default_desktop_environment.desktop_file,
                self.root_mount_point
                )
            )

    def greeter_setup(self):
        pass


class DMgdm(DisplayManager):
    name = "gdm"
    executable = "gdm"
    config = None  # Set by have_dm()

    def have_dm(self):
        """
        GDM exists with different executable names, so search
        for one of them and use it.
        """
        for executable, config in (
            ( "gdm", "etc/gdm/custom.conf" ),
            ( "gdm3", "etc/gdm3/daemon.conf" )
        ):
            bin_path = "{!s}/usr/bin/{!s}".format(self.root_mount_point, executable)
            sbin_path = "{!s}/usr/sbin/{!s}".format(self.root_mount_point, executable)
            if os.path.exists(bin_path) or os.path.exists(sbin_path):
                # Keep the found-executable name around for later
                self.executable = executable
                self.config = config
                return True

        return False

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        if self.config is None:
            raise ValueError( "No config file for GDM has been set." )

        # Systems with GDM as Desktop Manager
        gdm_conf_path = os.path.join(self.root_mount_point, self.config)

        if os.path.exists(gdm_conf_path):
            with open(gdm_conf_path, 'r') as gdm_conf:
                text = gdm_conf.readlines()

            with open(gdm_conf_path, 'w') as gdm_conf:
                for line in text:
                    if 'AutomaticLogin=' in line:
                        line = ""
                    if 'AutomaticLoginEnable=True' in line:
                        line = ""
                    if '[daemon]' in line:
                        if do_autologin:
                            line = (
                                "[daemon]\n"
                                "AutomaticLogin={!s}\n"
                                "AutomaticLoginEnable=True\n".format(username)
                                )
                        else:
                            line = "[daemon]\nAutomaticLoginEnable=False\n"

                    gdm_conf.write(line)
        else:
            with open(gdm_conf_path, 'w') as gdm_conf:
                gdm_conf.write(
                    '# Calamares - Enable automatic login for user\n'
                    )
                gdm_conf.write('[daemon]\n')

                if do_autologin:
                    gdm_conf.write("AutomaticLogin={!s}\n".format(username))
                    gdm_conf.write('AutomaticLoginEnable=True\n')
                else:
                    gdm_conf.write('AutomaticLoginEnable=False\n')

        if (do_autologin):
            accountservice_dir = "{!s}/var/lib/AccountsService/users".format(
                    self.root_mount_point
                    )
            userfile_path = "{!s}/{!s}".format(accountservice_dir, username)
            if os.path.exists(accountservice_dir):
                with open(userfile_path, "w") as userfile:
                    userfile.write("[User]\n")

                    if default_desktop_environment is not None:
                        userfile.write("XSession={!s}\n".format(
                            default_desktop_environment.desktop_file))

                    userfile.write("Icon=\n")


    def basic_setup(self):
        if libcalamares.utils.target_env_call(
                ['getent', 'group', 'gdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['groupadd', '-g', '120', 'gdm']
                )

        if libcalamares.utils.target_env_call(
                ['getent', 'passwd', 'gdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['useradd',
                    '-c', '"Gnome Display Manager"',
                    '-u', '120',
                    '-g', 'gdm',
                    '-d', '/var/lib/gdm',
                    '-s', '/usr/bin/nologin',
                    'gdm'
                    ]
                )

        libcalamares.utils.target_env_call(
            ['passwd', '-l', 'gdm']
            )
        libcalamares.utils.target_env_call(
            ['chown', '-R', 'gdm:gdm', '/var/lib/gdm']
            )

    def desktop_environment_setup(self, desktop_environment):
        pass

    def greeter_setup(self):
        pass


class DMkdm(DisplayManager):
    name = "kdm"
    executable = "kdm"

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        # Systems with KDM as Desktop Manager
        kdm_conf_path = os.path.join(
            self.root_mount_point, "usr/share/config/kdm/kdmrc"
            )
        # Check which path is in use: SUSE does something else.
        # Also double-check the default setting. Pick the first
        # one that exists in the target.
        for candidate_kdmrc in (
            "usr/share/config/kdm/kdmrc",
            "usr/share/kde4/config/kdm/kdmrc",
        ):
            p = os.path.join(self.root_mount_point, candidate_kdmrc)
            if os.path.exists(p):
                kdm_conf_path = p
                break
        text = []

        if os.path.exists(kdm_conf_path):
            with open(kdm_conf_path, 'r') as kdm_conf:
                text = kdm_conf.readlines()

            with open(kdm_conf_path, 'w') as kdm_conf:
                for line in text:
                    if 'AutoLoginEnable=' in line:
                        if do_autologin:
                            line = 'AutoLoginEnable=true\n'
                        else:
                            line = 'AutoLoginEnable=false\n'

                    if do_autologin and 'AutoLoginUser=' in line:
                        line = "AutoLoginUser={!s}\n".format(username)

                    kdm_conf.write(line)
        else:
            return (
                _("Cannot write KDM configuration file"),
                _("KDM config file {!s} does not exist").format(kdm_conf_path)
                )

    def basic_setup(self):
        if libcalamares.utils.target_env_call(
                ['getent', 'group', 'kdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['groupadd', '-g', '135', 'kdm']
                )

        if libcalamares.utils.target_env_call(
                ['getent', 'passwd', 'kdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['useradd',
                    '-u', '135',
                    '-g', 'kdm',
                    '-d', '/var/lib/kdm',
                    '-s', '/bin/false',
                    '-r',
                    '-M',
                    'kdm'
                    ]
                )

        libcalamares.utils.target_env_call(
            ['chown', '-R', '135:135', 'var/lib/kdm']
            )

    def desktop_environment_setup(self, desktop_environment):
        pass

    def greeter_setup(self):
        pass


class DMlxdm(DisplayManager):
    name = "lxdm"
    executable = "lxdm"

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        # Systems with LXDM as Desktop Manager
        lxdm_conf_path = os.path.join(self.root_mount_point, "etc/lxdm/lxdm.conf")
        text = []

        if os.path.exists(lxdm_conf_path):
            with open(lxdm_conf_path, 'r') as lxdm_conf:
                text = lxdm_conf.readlines()

            with open(lxdm_conf_path, 'w') as lxdm_conf:
                for line in text:
                    if 'autologin=' in line:
                        if do_autologin:
                            line = "autologin={!s}\n".format(username)
                        else:
                            line = "# autologin=\n"

                    lxdm_conf.write(line)
        else:
            return (
                _("Cannot write LXDM configuration file"),
                _("LXDM config file {!s} does not exist").format(lxdm_conf_path)
                )

    def basic_setup(self):
        if libcalamares.utils.target_env_call(
                ['getent', 'group', 'lxdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['groupadd', '--system', 'lxdm']
                )

        libcalamares.utils.target_env_call(
            ['chgrp', '-R', 'lxdm', '/var/lib/lxdm']
            )
        libcalamares.utils.target_env_call(
            ['chgrp', 'lxdm', '/etc/lxdm/lxdm.conf']
            )
        libcalamares.utils.target_env_call(
            ['chmod', '+r', '/etc/lxdm/lxdm.conf']
            )

    def desktop_environment_setup(self, default_desktop_environment):
        os.system(
            "sed -i -e \"s|^.*session=.*|session={!s}|\" "
            "{!s}/etc/lxdm/lxdm.conf".format(
                default_desktop_environment.executable,
                self.root_mount_point
                )
            )

    def greeter_setup(self):
        pass


class DMlightdm(DisplayManager):
    name = "lightdm"
    executable = "lightdm"

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        # Systems with LightDM as Desktop Manager
        # Ideally, we should use configparser for the ini conf file,
        # but we just do a simple text replacement for now, as it
        # worksforme(tm)
        lightdm_conf_path = os.path.join(
            self.root_mount_point, "etc/lightdm/lightdm.conf"
            )
        text = []
        addseat = False
        loopcount = 0

        if os.path.exists(lightdm_conf_path):
            with open(lightdm_conf_path, 'r') as lightdm_conf:
                text = lightdm_conf.readlines()
                # Check to make sure [SeatDefaults] or [Seat:*] is in the config,
                # otherwise we'll risk malforming the config
                addseat = '[SeatDefaults]' not in text and '[Seat:*]' not in text

            with open(lightdm_conf_path, 'w') as lightdm_conf:
                if addseat:
                    # Prepend Seat line to start of file rather than leaving it without one
                    # This keeps the config from being malformed for LightDM
                    text = ["[Seat:*]\n"] + text
                for line in text:
                    if 'autologin-user=' in line:
                        if do_autologin:
                            line = "autologin-user={!s}\n".format(username)
                        else:
                            line = "#autologin-user=\n"

                    lightdm_conf.write(line)
        else:
            try:
                # Create a new lightdm.conf file; this is documented to be
                # read last, after everything in lightdm.conf.d/
                with open(lightdm_conf_path, 'w') as lightdm_conf:
                    if do_autologin:
                        lightdm_conf.write(
                            "[Seat:*]\nautologin-user={!s}\n".format(username))
                    else:
                        lightdm_conf.write(
                            "[Seat:*]\n#autologin-user=\n")
            except FileNotFoundError:
                return (
                    _("Cannot write LightDM configuration file"),
                    _("LightDM config file {!s} does not exist").format(lightdm_conf_path)
                    )


    def basic_setup(self):
        libcalamares.utils.target_env_call(
            ['mkdir', '-p', '/run/lightdm']
            )

        if libcalamares.utils.target_env_call(
                ['getent', 'group', 'lightdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['groupadd', '-g', '620', 'lightdm']
                )

        if libcalamares.utils.target_env_call(
                ['getent', 'passwd', 'lightdm']
                ) != 0:
            libcalamares.utils.target_env_call(
                ['useradd', '-c',
                    '"LightDM Display Manager"',
                    '-u', '620',
                    '-g', 'lightdm',
                    '-d', '/var/run/lightdm',
                    '-s', '/usr/bin/nologin',
                    'lightdm'
                    ]
                )

        libcalamares.utils.target_env_call(['passwd', '-l', 'lightdm'])
        libcalamares.utils.target_env_call(['chown', '-R', 'lightdm:lightdm', '/run/lightdm'])
        libcalamares.utils.target_env_call(['chmod', '+r' '/etc/lightdm/lightdm.conf'])

    def desktop_environment_setup(self, default_desktop_environment):
        os.system(
            "sed -i -e \"s/^.*user-session=.*/user-session={!s}/\" "
            "{!s}/etc/lightdm/lightdm.conf".format(
                default_desktop_environment.desktop_file,
                self.root_mount_point
                )
            )

    def greeter_setup(self):
        lightdm_conf_path = os.path.join(
            self.root_mount_point, "etc/lightdm/lightdm.conf"
            )

        # configure lightdm-greeter
        greeter_path = os.path.join(
            self.root_mount_point, "usr/share/xgreeters"
            )

        if (os.path.exists(greeter_path)):
            # configure first found lightdm-greeter
            for entry in os.listdir(greeter_path):
                if entry.endswith('.desktop'):
                    greeter = entry.split('.')[0]
                    libcalamares.utils.debug(
                        "found greeter {!s}".format(greeter)
                        )
                    os.system(
                        "sed -i -e \"s/^.*greeter-session=.*"
                        "/greeter-session={!s}/\" {!s}".format(
                            greeter,
                            lightdm_conf_path
                            )
                        )
                    libcalamares.utils.debug(
                        "{!s} configured as greeter.".format(greeter)
                        )
                    break
            else:
                return (
                    _("Cannot configure LightDM"),
                    _("No LightDM greeter installed.")
                    )


class DMslim(DisplayManager):
    name = "slim"
    executable = "slim"

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        # Systems with Slim as Desktop Manager
        slim_conf_path = os.path.join(self.root_mount_point, "etc/slim.conf")
        text = []

        if os.path.exists(slim_conf_path):
            with open(slim_conf_path, 'r') as slim_conf:
                text = slim_conf.readlines()

            with open(slim_conf_path, 'w') as slim_conf:
                for line in text:
                    if 'auto_login' in line:
                        if do_autologin:
                            line = 'auto_login yes\n'
                        else:
                            line = 'auto_login no\n'

                    if do_autologin and 'default_user' in line:
                        line = "default_user {!s}\n".format(username)

                    slim_conf.write(line)
        else:
            return (
                _("Cannot write SLIM configuration file"),
                _("SLIM config file {!s} does not exist").format(slim_conf_path)
                )


    def basic_setup(self):
        pass

    def desktop_environment_setup(self, desktop_environment):
        pass

    def greeter_setup(self):
        pass


class DMsddm(DisplayManager):
    name = "sddm"
    executable = "sddm"

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        import configparser

        # Systems with Sddm as Desktop Manager
        sddm_conf_path = os.path.join(self.root_mount_point, "etc/sddm.conf")

        sddm_config = configparser.ConfigParser(strict=False)
        # Make everything case sensitive
        sddm_config.optionxform = str

        if os.path.isfile(sddm_conf_path):
            sddm_config.read(sddm_conf_path)

        if 'Autologin' not in sddm_config:
            sddm_config.add_section('Autologin')

        if do_autologin:
            sddm_config.set('Autologin', 'User', username)
        elif sddm_config.has_option('Autologin', 'User'):
            sddm_config.remove_option('Autologin', 'User')

        if default_desktop_environment is not None:
            sddm_config.set(
                'Autologin',
                'Session',
                default_desktop_environment.desktop_file
                )

        with open(sddm_conf_path, 'w') as sddm_config_file:
            sddm_config.write(sddm_config_file, space_around_delimiters=False)


    def basic_setup(self):
        pass

    def desktop_environment_setup(self, desktop_environment):
        pass

    def greeter_setup(self):
        pass


class DMgreetd(DisplayManager):
    name = "greetd"
    executable = "greetd"
    greeter_user = "greeter"
    greeter_group = "greetd"
    config_data = {}

    def os_path(self, path):
        return os.path.join(self.root_mount_point, path)

    def config_path(self):
        return self.os_path("etc/greetd/config.toml")

    def environments_path(self):
        return self.os_path("etc/greetd/environments")

    def config_load(self):
        import toml

        if (os.path.exists(self.config_path())):
            with open(self.config_path(), "r") as f:
                self.config_data = toml.load(f)

        self.config_data['terminal'] = dict(vt = "next")

        default_session_group = self.config_data.get('default_session', None)
        if not default_session_group:
            self.config_data['default_session'] = {}

        self.config_data['default_session']['user'] = self.greeter_user

        return self.config_data

    def config_write(self):
        import toml
        with open(self.config_path(), "w") as f:
            toml.dump(self.config_data, f)

    def basic_setup(self):
        if libcalamares.utils.target_env_call(
                ['getent', 'group', self.greeter_group]
                ) != 0:
            libcalamares.utils.target_env_call(
                ['groupadd', self.greeter_group]
                )

        if libcalamares.utils.target_env_call(
                ['getent', 'passwd', self.greeter_user]
                ) != 0:
            libcalamares.utils.target_env_call(
                ['useradd',
                    '-c', '"Greeter User"',
                    '-g', self.greeter_group,
                    '-s', '/bin/bash',
                    self.greeter_user
                    ]
                )

    def desktop_environment_setup(self, default_desktop_environment):
        with open(self.environments_path(), 'w') as envs_file:
            envs_file.write(default_desktop_environment.desktop_file)
            envs_file.write("\n")

    def greeter_setup(self):
        pass

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        self.config_load()

        de_command = default_desktop_environment.executable
        if os.path.exists(self.os_path("usr/bin/gtkgreed")) and os.path.exists(self.os_path("usr/bin/cage")):
            self.config_data['default_session']['command'] = "cage -s -- gtkgreet"
        elif os.path.exists(self.os_path("usr/bin/tuigreet")):
            tuigreet_base_cmd = "tuigreet --remember --time --issue --asterisks --cmd "
            self.config_data['default_session']['command'] = tuigreet_base_cmd + de_command
        elif os.path.exists(self.os_path("usr/bin/ddlm")):
            self.config_data['default_session']['command'] = "ddlm --target " + de_command
        else:
            self.config_data['default_session']['command'] = "agreety --cmd " + de_command

        if do_autologin == True:
            self.config_data['initial_session'] = dict(command = de_command, user = username)

        self.config_write()


class DMsysconfig(DisplayManager):
    name = "sysconfig"
    executable = None

    def set_autologin(self, username, do_autologin, default_desktop_environment):
        dmauto = "DISPLAYMANAGER_AUTOLOGIN"

        os.system(
            "sed -i -e 's|^{!s}=.*|{!s}=\"{!s}\"|' "
            "{!s}/etc/sysconfig/displaymanager".format(
                            dmauto, dmauto,
                            username if do_autologin else "",
                            self.root_mount_point
                            )
            )


    def basic_setup(self):
        pass

    def desktop_environment_setup(self, desktop_environment):
        pass

    def greeter_setup(self):
        pass

    # For openSUSE-derivatives, there is only sysconfig to configure,
    # and no special DM configuration for it. Instead, check that
    # sysconfig is available in the target.
    def have_dm(self):
        config = "{!s}/etc/sysconfig/displaymanager".format(self.root_mount_point)
        return os.path.exists(config)


# Collect all the subclasses of DisplayManager defined above,
# and index them based on the name property of each class.
display_managers = [
    (c.name, c)
    for c in globals().values()
    if type(c) is abc.ABCMeta and issubclass(c, DisplayManager) and c.name
]


def run():
    """
    Configure display managers.

    We acquire a list of displaymanagers, either from config or (overridden)
    from globalstorage. This module will try to set up (including autologin)
    all the displaymanagers in the list, in that specific order. Most distros
    will probably only ship one displaymanager.
    If a displaymanager is in the list but not installed, a debugging message
    is printed and the entry ignored.
    """
    # Get configuration settings for display managers
    displaymanagers = None
    if "displaymanagers" in libcalamares.job.configuration:
        displaymanagers = libcalamares.job.configuration["displaymanagers"]

    if libcalamares.globalstorage.contains("displayManagers"):
        displaymanagers = libcalamares.globalstorage.value("displayManagers")

    if ("sysconfigSetup" in libcalamares.job.configuration
            and libcalamares.job.configuration["sysconfigSetup"]):
        displaymanagers = ["sysconfig"]

    if not displaymanagers:
        return (
            _("No display managers selected for the displaymanager module."),
            _("The displaymanagers list is empty or undefined in both "
            "globalstorage and displaymanager.conf.")
            )

    # Get instances that are actually installed
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")
    dm_impl = []
    dm_names = displaymanagers[:]
    for dm in dm_names:
        # Find the implementation class
        dm_instance = None
        impl = [ cls for name, cls in display_managers if name == dm ]
        if len(impl) == 1:
            dm_instance = impl[0](root_mount_point)
            if dm_instance.have_dm():
                dm_impl.append(dm_instance)
            else:
                dm_instance = None
        else:
            libcalamares.utils.debug("{!s} has {!s} implementation classes.".format(dm, len(impl)))

        if dm_instance is None:
            libcalamares.utils.debug("{!s} selected but not installed".format(dm))
            if dm in displaymanagers:
                displaymanagers.remove(dm)

    if not dm_impl:
        libcalamares.utils.warning(
            "No display managers selected for the displaymanager module. "
            "The list is empty after checking for installed display managers."
            )
        return None

    # Pick up remaining settings
    if "defaultDesktopEnvironment" in libcalamares.job.configuration:
        entry = libcalamares.job.configuration["defaultDesktopEnvironment"]
        default_desktop_environment = DesktopEnvironment(
            entry["executable"], entry["desktopFile"]
            )
        # Adjust if executable is bad, but desktopFile isn't.
        if not default_desktop_environment.update_from_desktop_file(root_mount_point):
            libcalamares.utils.warning(
                "The configured default desktop environment, {!s}, "
                "can not be found.".format(default_desktop_environment.desktop_file))
    else:
        default_desktop_environment = find_desktop_environment(
            root_mount_point
            )

    if "basicSetup" in libcalamares.job.configuration:
        enable_basic_setup = libcalamares.job.configuration["basicSetup"]
    else:
        enable_basic_setup = False

    username = libcalamares.globalstorage.value("autoLoginUser")
    if username is not None:
        do_autologin = True
        libcalamares.utils.debug("Setting up autologin for user {!s}.".format(username))
    else:
        do_autologin = False
        libcalamares.utils.debug("Unsetting autologin.")

    libcalamares.globalstorage.insert("displayManagers", displaymanagers)

    # Do the actual configuration and collect messages
    dm_setup_message = []
    for dm in dm_impl:
        dm_message = None
        if enable_basic_setup:
            dm_message = dm.basic_setup()
        if default_desktop_environment is not None and dm_message is None:
            dm_message = dm.desktop_environment_setup(default_desktop_environment)
        if dm_message is None:
            dm_message = dm.greeter_setup()
        if dm_message is None:
            dm_message = dm.set_autologin(username, do_autologin, default_desktop_environment)

        if dm_message is not None:
            dm_setup_message.append("{!s}: {!s}".format(*dm_message))

    if dm_setup_message:
        return (
            _("Display manager configuration was incomplete"),
            "\n".join(dm_setup_message)
            )



#   SPDX-FileCopyrightText: no
#   SPDX-License-Identifier: CC0-1.0
#
# Stubs for part of the Python API from libcalamares
# (although the **actual** API is presented through
# Boost::Python, not as a bare C-extension) so that
# pylint doesn't complain about libcalamares internals.

configuration = dict()

def setprogress(_): pass

def pretty_name(): return ""

def working_path(): return ""



#! /usr/bin/env python3
#
#   SPDX-FileCopyrightText: 2018 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: BSD-2-Clause
#
# Uses the Transifex API to get a list of enabled languages,
# and outputs CMake settings for inclusion into CMakeLists.txt.
#
# This is a Python3 script.
#
# Run it with a -v command-line option to get extra output on
# actual translation percentages.
import sys
import os
import argparse

class TXError(Exception):
    pass


class TransifexGetter(object):
    """
    Get language data from Transifex.

    The object does all the work in __init__, after that
    the only relevant data is .languages, a dictionary
    of language data.
    """
    def __init__(self):
        token = self.get_tx_credentials()
        if token is None:
            raise TXError("Could not get Transifex API token")

        import requests
        r = requests.get("https://api.transifex.com/organizations/calamares/projects/calamares/resources/calamares/", auth=("api", token))
        if r.status_code != 200:
            raise TXError("Could not get Transifex data from API")

        j = r.json()
        self.languages = j["stats"]


    def get_tx_credentials(self):
        """
        Gets the API token out of the user's .transifexrc (this is supposed
        to be secure).
        """
        import configparser
        import os
        txconfig_name = os.path.expanduser("~/.transifexrc")
        try:
            with open(txconfig_name, "r") as f:
                parser = configparser.ConfigParser()
                parser.read_file(f)

                return parser.get("https://www.transifex.com", "password")
        except IOError as e:
            return None


class BogusGetter(object):
    """
    Fake language data.

    This object pretends to retrieve data, and returns fixed language lists and percentages,
    for testing purposes without hitting Transifex servers all the time.
    """
    def __init__(self):
        self.languages = dict()
        for lang, completion in ( ("sq", 100), ("ar", 44), ("as", 28), ("de", 15), ("da", 4), ("ts", 82) ):
            self.languages[lang] = dict(translated=dict(stringcount=686, percentage=(completion/100.0)))


class PrintOutputter(object):
    """
    Output via print-statements.
    """
    def __init__(self):
        pass

    def print(self, s):
        print(s)

    def __enter__(self):
        return self

    def __exit__(self, e, v, tb):
        pass


class EditingOutputter(object):
    """
    Edit CMakeLists in-place.
    """
    def __init__(self):
        with open("CMakeLists.txt", "r") as f:
            lines = f.readlines()

        mark = None
        for l in lines:
            # Note that we didn't strip the lines, so need the \n here
            if l.startswith("# Total ") and l.endswith(" languages\n"):
                mark = lines.index(l)
                break
        if mark is None:
            raise TXError("No CMakeLists.txt lines for TX stats found")
        self.pre_lines = lines[:mark]

        nextmark = mark + 1
        for l in lines[mark+1:]:
            if l.startswith("set( _tx_"):
                nextmark += 1
                continue
            if l.startswith("    "):
                nextmark += 1
                continue
            break
        if nextmark > mark + 12 or nextmark > len(lines) - 4:
            # Try to catch runaway nextmarks: we know there should
            # be four set-lines, which are unlikely to be 3 lines each;
            # similarly the CMakeLists.txt is supposed to end with
            # some boilerplate.
            raise TXError("Could not find end of TX settings in CMakeLists.txt")
        self.post_lines = lines[nextmark:]

        self.mid_lines = []
        print("# Editing CMakeLists.txt in-place")

    def print(self, s):
        # Add the implicit \n from print()
        self.mid_lines.append(s + "\n")
        if s.startswith("#"):
            print(s)

    def __enter__(self):
        return self

    def __exit__(self, e, v, tb):
        if e is None:
            with open("CMakeLists.txt", "w") as f:
                f.write("".join(self.pre_lines + self.mid_lines + self.post_lines))
            print("# CMakeLists.txt updated")


def output_langs(all_langs, outputter, label, filterfunc):
    """
    Output (via print) all of the languages in @p all_langs
    that satisfy the translation-percentage filter @p filterfunc.
    Prints a CMake set() command with the @p label as part
    of the variable name.

    Performs line-wrapping.
    """
    these_langs = [l for s, l in all_langs if filterfunc(s)]
    out = " ".join(["set( _tx_%s" % label, " ".join(sorted(these_langs)), ")"])
    width = 68
    prefix = ""

    while len(out) > width - len(prefix):
        chunk = out[:out[:width].rfind(" ")]
        outputter.print("%s%s" % (prefix, chunk))
        out = out[len(chunk)+1:]
        prefix = "    "
    outputter.print("%s%s" % (prefix, out))


def get_tx_stats(languages, outputter, verbose):
    """
    Does an API request to Transifex with the given API @p token, getting
    the translation statistics for the main body of texts. Then prints
    out CMake settings to replace the _tx_* variables in CMakeLists.txt
    according to standard criteria.

    If @p verbose is True, prints out language stats as well.
    """
    # Some languages go into the "incomplete" list by definition,
    # regardless of their completion status: this can have various reasons.
    #
    # Note that Esperanto (eo) is special-cased in CMakeLists.txt
    # during the build; recent Qt releases *do* support the language,
    # and it's at-the-least ok.
    incomplete_languages = (
        "eo",   # Not supported by QLocale < 5.12.1
        "ie",   # Not supported by Qt at least through 5.15.0
        )

    all_langs = []
    outputter.print("# Total %d languages" % len(languages))
    for lang_name in languages:
        stats = languages[lang_name]["translated"]["percentage"]
        # Make the by-definition-incomplete languages have a percentage
        # lower than zero; this way they end up sorted (in -v output)
        # at the bottom but you can still determine the "actual" percentage.
        if lang_name in incomplete_languages:
            stats = -stats
        all_langs.append((stats, lang_name))

    if verbose:
        for s, l in sorted(all_langs, reverse=True):
            outputter.print("#  %16s\t%6.2f" % (l, s * 100.0))
    output_langs(all_langs, outputter, "complete", lambda s : s == 1.0)
    output_langs(all_langs, outputter, "good", lambda s : 1.0 > s >= 0.75)
    output_langs(all_langs, outputter, "ok", lambda s : 0.75 > s >= 0.05)
    output_langs(all_langs, outputter, "incomplete", lambda s : 0.05 > s)

    # Audit the languages that are in TX, mapped to git
    for lang_name in languages:
        if not os.path.exists("lang/calamares_{}.ts".format(lang_name)):
            print("# !! Missing translation file for {}".format(lang_name))
        if not os.path.isdir("lang/python/{}/LC_MESSAGES".format(lang_name)):
            print("# !! Missing Python translation file for {}".format(lang_name))

    # Audit the files that are in git, mapped to TX
    special_cases = ("python.pot", "python", "CMakeLists.txt", "txload.cpp", "calamares_i18n.qrc.in")
    for file_name in os.listdir("lang"):
        if file_name in special_cases:
            continue
        elif file_name.startswith("calamares_") and file_name.endswith(".ts"):
            key = file_name[10:-3]
            if not key in languages and not key == "en":
                print("# !! Translation file for {} not in TX".format(key))
        elif file_name.startswith("tz_") and file_name.endswith(".ts"):
            key = file_name[3:-3]
            if not key in languages and not key == "en":
                print("# !! Translation file for TZ {} not in TX".format(key))
        elif file_name.startswith("kb_") and file_name.endswith(".ts"):
            key = file_name[3:-3]
            if not key in languages and not key == "en":
                print("# !! Translation file for KB {} not in TX".format(key))
        else:
            print("# !! Weird translation file {} not in TX".format(file_name))

    # Audit the python translation files that are in git, mapped to TX
    for file_name in os.listdir("lang/python"):
        if file_name not in languages:
            print("# !! Translation file for Python {} not in TX".format(file_name))

    return 0


def main():
    parser = argparse.ArgumentParser(description="Update Transifex Statistics")
    parser.add_argument("--verbose", "-v", help="Show statistics", action="store_true")
    parser.add_argument("--bogus", "-n", help="Use bogus data (do not query Transifex)", action="store_true")
    parser.add_argument("--edit", "-e", help="Edit CMakeLists.txt in-place", action="store_true")
    args = parser.parse_args()
    try:
        if args.bogus:
            getter = BogusGetter()
        else:
            getter = TransifexGetter()
        if args.edit:
            outputter = EditingOutputter()
        else:
            outputter = PrintOutputter()
        with outputter:
            return get_tx_stats(getter.languages, outputter, args.verbose)
    except TXError as e:
        print("! " + str(e))
        return 1;
    return 0

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-FileCopyrightText: 2021 Anke boersma <demm@koasx.us>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import os
import shutil

import libcalamares

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Saving network configuration.")


def get_live_user():
    """
    Gets the "live user" login. This might be "live", or "nitrux",
    or something similar: it is the login name used *right now*,
    and network configurations saved for that user, should be applied
    also for the installed user (which probably has a different name).
    """
    # getlogin() is a thin-wrapper, and depends on getlogin(3),
    # which reads utmp -- and utmp isn't always set up right.
    try:
        return os.getlogin()
    except OSError:
        pass
    # getpass will return the **current** user, which is generally root.
    # That isn't very useful, because the network settings have been
    # made outside of Calamares-running-as-root, as a different user.
    #
    # If Calamares is running as non-root, though, this is fine.
    import getpass
    name = getpass.getuser()
    if name != "root":
        return name

    # TODO: other mechanisms, e.g. guessing that "live" is the name
    # TODO: support a what-is-the-live-user setting
    return None


def replace_username(nm_config_filename, live_user, target_user):
    """
    If @p live_user isn't None, then go through the given
    file and replace @p live_user by the @p target_user.

    Reads the file, then (re-)writes it with new permissions lives.
    """
    # FIXME: Perhaps if live_user is None, we should just replace **all**
    #        permissions lines? After all, this is supposed to be a live
    #        system so **whatever** NM networks are configured, should be
    #        available to the new user.
    if live_user is None:
        return
    if not os.path.exists(nm_config_filename):
        return

    with open(nm_config_filename, "r", encoding="UTF-8") as network_conf:
        text = network_conf.readlines()

    live_permissions = 'permissions=user:{}:;'.format(live_user)
    target_permissions = 'permissions=user:{}:;\n'.format(target_user)
    with open(nm_config_filename, "w", encoding="UTF-8") as network_conf:
        for line in text:
            if live_permissions in line:
                line = target_permissions
            network_conf.write(line)


def path_pair(root_mount_point, relative_path):
    """
    Returns /relative_path and the relative path in the target system.
    """
    return ("/" + relative_path, os.path.join(root_mount_point, relative_path))


def run():
    """
    Setup network configuration
    """
    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")
    user = libcalamares.globalstorage.value("username")
    live_user = get_live_user()

    if root_mount_point is None:
        libcalamares.utils.warning("rootMountPoint is empty, {!s}".format(root_mount_point))
        return (_("Configuration Error"),
                _("No root mount point is given for <pre>{!s}</pre> to use." ).format("networkcfg"))

    source_nm, target_nm = path_pair(root_mount_point, "etc/NetworkManager/system-connections/")

    # Sanity checks.  We don't want to do anything if a network
    # configuration already exists on the target
    if os.path.exists(source_nm) and os.path.exists(target_nm):
        for network in os.listdir(source_nm):
            # Skip LTSP live
            if network == "LTSP":
                continue

            source_network = os.path.join(source_nm, network)
            target_network = os.path.join(target_nm, network)

            if os.path.exists(target_network):
                continue

            try:
                shutil.copy(source_network, target_network, follow_symlinks=False)
                replace_username(target_network, live_user, user)
            except FileNotFoundError:
                libcalamares.utils.debug(
                    "Can't copy network configuration files in {}".format(source_network)
                    )
            except FileExistsError:
                pass

    # We need to overwrite the default resolv.conf in the chroot.
    source_resolv, target_resolv = path_pair(root_mount_point, "etc/resolv.conf")
    if source_resolv != target_resolv and os.path.exists(source_resolv):
        try:
            os.remove(target_resolv)
        except Exception as err:
            libcalamares.utils.debug(
                "Couldn't remove {}: {}".format(target_resolv, err)
                )

        try:
            shutil.copy(source_resolv, target_resolv, follow_symlinks=False)
        except Exception as err:
            libcalamares.utils.debug(
                "Can't copy resolv.conf from {}: {}".format(source_resolv, err)
                )

    return None



#   SPDX-FileCopyrightText: no
#   SPDX-License-Identifier: CC0-1.0
#
# Calamares Boilerplate
import libcalamares
libcalamares.globalstorage = libcalamares.GlobalStorage(None)
libcalamares.globalstorage.insert("testing", True)

# Module prep-work
from src.modules.packages import main

# .. we don't have a job in this test, so fake one
class Job(object):
    def __init__(self, filename):
        self.configuration = libcalamares.utils.load_yaml(filename) if filename is not None else dict()

import sys
if len(sys.argv) > 4:
    filename = sys.argv[1]
    retry = int(sys.argv[2])
    timeout = bool(int(sys.argv[3]))
    needed = bool(int(sys.argv[4]))
else:
    filename = None
    retry = 0
    timeout = False
    needed = False

libcalamares.utils.warning("Expecting {!s} retry={!s} timeout={!s} needed={!s}".format(filename, retry, timeout, needed))

# Specific PM test
libcalamares.job = Job(filename)
p = main.PMPacman()
assert p.pacman_num_retries == retry, "{!r} vs {!r}".format(p.pacman_num_retries, retry)
assert p.pacman_disable_timeout == timeout, "{!r} vs {!r}".format(p.pacman_disable_timeout, timeout)
assert p.pacman_needed_only == needed, "{!r} vs {!r}".format(p.pacman_needed_only, needed)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2017 Ghiunhan Mamut <venerix@redcorelinux.org>
#   SPDX-FileCopyrightText: 2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import os.path

import libcalamares

import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext



def pretty_name():
    return _("Configuring OpenRC dmcrypt service.")


def write_dmcrypt_conf(partitions, root_mount_point, dmcrypt_conf_path):
    crypto_target = ""
    crypto_source = ""
    unencrypted_separate_boot = any(p["mountPoint"] == "/boot" and "luksMapperName" not in p for p in partitions)

    for partition in partitions:
        has_luks = "luksMapperName" in partition
        skip_partitions = partition["mountPoint"] == "/" or partition["fs"] == "linuxswap"

        if not has_luks and not skip_partitions:
            libcalamares.utils.debug(
                "Skip writing OpenRC LUKS configuration for partition {!s}".format(partition["mountPoint"]))
        if has_luks and not skip_partitions:
            crypto_target = partition["luksMapperName"]
            crypto_source = "/dev/disk/by-uuid/{!s}".format(partition["uuid"])
            libcalamares.utils.debug(
                "Writing OpenRC LUKS configuration for partition {!s}".format(partition["mountPoint"]))

            with open(os.path.join(root_mount_point, dmcrypt_conf_path), 'a+') as dmcrypt_file:
                dmcrypt_file.write("\ntarget=" + crypto_target)
                dmcrypt_file.write("\nsource=" + crypto_source)
                # Don't use keyfile if boot is unencrypted, keys must not be stored on unencrypted partitions
                if not unencrypted_separate_boot:
                    dmcrypt_file.write("\nkey=/crypto_keyfile.bin")
                dmcrypt_file.write("\n")

        if has_luks and skip_partitions:
            pass  # root and swap partitions should be handled by initramfs generators

    return None

def run():
    """
    This module configures OpenRC dmcrypt service for LUKS encrypted partitions.
    :return:
    """

    root_mount_point = libcalamares.globalstorage.value("rootMountPoint")
    dmcrypt_conf_path = libcalamares.job.configuration["configFilePath"]
    partitions = libcalamares.globalstorage.value("partitions")

    if not partitions:
        libcalamares.utils.warning("partitions is empty, {!s}".format(partitions))
        return (_("Configuration Error"),
                _("No partitions are defined for <pre>{!s}</pre> to use." ).format("openrcdmcryptcfg"))
    if not root_mount_point:
        libcalamares.utils.warning("rootMountPoint is empty, {!s}".format(root_mount_point))
        return (_("Configuration Error"),
                _("No root mount point is given for <pre>{!s}</pre> to use." ).format("openrcdmcryptcfg"))

    dmcrypt_conf_path = dmcrypt_conf_path.lstrip('/')

    return write_dmcrypt_conf(partitions, root_mount_point, dmcrypt_conf_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2014 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2014 Teo Mrnjavac <teo@kde.org>
#   SPDX-FileCopyrightText: 2017 Alf Gaida <agaida@siduction.org>
#   SPDX-FileCopyrightText: 2018-2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares


import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configure systemd services")


def systemctl(targets, command, suffix):
    """
    For each entry in @p targets, run "systemctl <command> <thing>",
    where <thing> is the entry's name plus the given @p suffix.
    (No dot is added between name and suffix; suffix may be empty)

    Returns a failure message, or None if this was successful.
    Services that are not mandatory have their failures suppressed
    silently.
    """
    for svc in targets:
        if isinstance(svc, str):
            name = svc
            mandatory = False
        else:
            name = svc["name"]
            mandatory = svc.get("mandatory", False)

        ec = libcalamares.utils.target_env_call(
            ['systemctl', command, "{}{}".format(name, suffix)]
            )

        if ec != 0:
            libcalamares.utils.warning(
                "Cannot {} systemd {} {}".format(command, suffix, name)
                )
            libcalamares.utils.warning(
                "systemctl {} call in chroot returned error code {}".format(command, ec)
                )
            if mandatory:
                title = _("Cannot modify service")
                diagnostic = _("<code>systemctl {arg!s}</code> call in chroot returned error code {num!s}.").format(arg=command, num=ec)

                if command == "enable" and suffix == ".service":
                    description = _("Cannot enable systemd service <code>{name!s}</code>.")
                elif command == "enable" and suffix == ".target":
                    description = _("Cannot enable systemd target <code>{name!s}</code>.")
                elif command == "enable" and suffix == ".timer":
                    description = _("Cannot enable systemd timer <code>{name!s}</code>.")
                elif command == "disable" and suffix == ".service":
                    description = _("Cannot enable systemd service <code>{name!s}</code>.")
                elif command == "disable" and suffix == ".target":
                    description = _("Cannot disable systemd target <code>{name!s}</code>.")
                elif command == "mask":
                    description = _("Cannot mask systemd unit <code>{name!s}</code>.")
                else:
                    description = _("Unknown systemd commands <code>{command!s}</code> and <code>{suffix!s}</code> for unit {name!s}.")

                return (title,
                        description.format(name=name, command=command, suffix=suffix) + " " + diagnostic
                        )
    return None


def run():
    """
    Setup systemd services
    """
    cfg = libcalamares.job.configuration

    # note that the "systemctl enable" and "systemctl disable" commands used
    # here will work in a chroot; in fact, they are the only systemctl commands
    # that support that, see:
    # http://0pointer.de/blog/projects/changing-roots.html

    r = systemctl(cfg.get("services", []), "enable", ".service")
    if r is not None:
        return r

    r = systemctl(cfg.get("targets", []), "enable", ".target")
    if r is not None:
        return r

    r = systemctl(cfg.get("timers", []), "enable", ".timer")
    if r is not None:
        return r

    r = systemctl(cfg.get("disable", []), "disable", ".service")
    if r is not None:
        return r

    r = systemctl(cfg.get("disable-targets", []), "disable", ".target")
    if r is not None:
        return r

    r = systemctl(cfg.get("mask", []), "mask", "")
    if r is not None:
        return r

    return None



#! /usr/bin/env python3
#
#   SPDX-FileCopyrightText: 2020 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: BSD-2-Clause
#
# Reduce a translation file -- generally, a Timezone translation -- by
# dropping untranslated strings. An untranslated string is one that
# has an empty translation **and** is marked unfinished.
#
# This is mostly useful to cut down the size of the source file:
# far and away most of the zones are not translated, and it's just a
# handful of places that get special treatment.

from xml.dom.minidom import parse
import sys

valid = True
dom = parse(sys.argv[1])
for n in dom.getElementsByTagName("translation"):
    attrs = n.attributes.keys()

    drop = True
    if "type" not in attrs:
        drop = False
    elif "type" in attrs and n.attributes["type"].value != "unfinished":
        # In the samples I've seen, only "unfinished" is a valid type;
        # once something has been translated, the attribute vanishes (see
        # the if branch, above).
        print("WARNING ''{!s}'' unknown type".format(n.attributes["type"].value))
        drop = False
        valid = False
    else:
        t = n.firstChild
        if t is None:
            # Unfinished and empty
            drop = True
        else:
            drop = bool(t.data)
    if drop:
        message = n.parentNode
        message.parentNode.removeChild(message)
        message.unlink()

if valid:
    for line in dom.toxml().split("\n"):
        if line.strip():
            print(line)



#   SPDX-FileCopyrightText: no
#   SPDX-License-Identifier: CC0-1.0
#
# Calamares Boilerplate
import libcalamares
libcalamares.globalstorage = libcalamares.GlobalStorage(None)
libcalamares.globalstorage.insert("testing", True)

# Module prep-work
from src.modules.displaymanager import main
default_desktop_environment = main.DesktopEnvironment("startplasma-x11", "kde-plasma.desktop")

import os
os.makedirs("/tmp/etc/greetd/", exist_ok=True)
try:
    os.remove("/tmp/etc/greetd/config.toml")
except FileNotFoundError as e:
    pass

try:
    import toml
except ImportError:
    # This is a failure of the test-environment.
    import sys
    print("Can't find module toml.", file=sys.stderr)
    sys.exit(0)

# Specific DM test
d = main.DMgreetd("/tmp")
d.set_autologin("d", True, default_desktop_environment)
# .. and again (this time checks load/save)
d.set_autologin("d", True, default_desktop_environment)
d.set_autologin("d", True, default_desktop_environment)



#   SPDX-FileCopyrightText: no
#   SPDX-License-Identifier: CC0-1.0
#
# Stubs for part of the Python API from libcalamares
# (although the **actual** API is presented through
# Boost::Python, not as a bare C-extension) so that
# pylint doesn't complain about libcalamares internals.

def debug(_): pass

def warning(_): pass

def error(_): pass

def gettext_path(): pass

def gettext_languages(): pass

def target_env_call(_): return 0

def check_target_env_call(_): pass

def target_env_process_output(cmd, *args): return 0

def host_env_process_output(cmd, *args): return 0

def mount(device, mountpoint, fstype, options): return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# === This file is part of Calamares - <https://calamares.io> ===
#
#   SPDX-FileCopyrightText: 2016 Artoo <artoo@manjaro.org>
#   SPDX-FileCopyrightText: 2017 Philip Müller <philm@manjaro.org>
#   SPDX-FileCopyrightText: 2018 Artoo <artoo@artixlinux.org>
#   SPDX-FileCopyrightText: 2018-2019 Adriaan de Groot <groot@kde.org>
#   SPDX-License-Identifier: GPL-3.0-or-later
#
#   Calamares is Free Software: see the License-Identifier above.
#

import libcalamares

from libcalamares.utils import target_env_call, warning
from os.path import exists, join


import gettext
_ = gettext.translation("calamares-python",
                        localedir=libcalamares.utils.gettext_path(),
                        languages=libcalamares.utils.gettext_languages(),
                        fallback=True).gettext


def pretty_name():
    return _("Configure OpenRC services")


class OpenrcController:
    """
    This is the openrc service controller.
    All of its state comes from global storage and the job
    configuration at initialization time.
    """

    def __init__(self):
        self.root = libcalamares.globalstorage.value('rootMountPoint')

        # Translate the entries in the config to the actions passed to rc-config
        self.services = dict()
        self.services["add"] = libcalamares.job.configuration.get('services', [])
        self.services["del"] = libcalamares.job.configuration.get('disable', [])

        self.initdDir = libcalamares.job.configuration['initdDir']
        self.runlevelsDir = libcalamares.job.configuration['runlevelsDir']


    def make_failure_description(self, state, name, runlevel):
        """
        Returns a generic "could not <foo>" failure message, specialized
        for the action @p state and the specific service @p name in @p runlevel.
        """
        if state == "add":
            description = _("Cannot add service {name!s} to run-level {level!s}.")
        elif state == "del":
            description = _("Cannot remove service {name!s} from run-level {level!s}.")
        else:
            description = _("Unknown service-action <code>{arg!s}</code> for service {name!s} in run-level {level!s}.")

        return description.format(arg=state, name=name, level=runlevel)


    def update(self, state):
        """
        Call rc-update for each service listed
        in services for the given @p state. rc-update
        is called with @p state as the command as well.
        """

        for svc in self.services.get(state, []):
            if isinstance(svc, str):
                name = svc
                runlevel = "default"
                mandatory = False
            else:
                name = svc["name"]
                runlevel = svc.get("runlevel", "default")
                mandatory = svc.get("mandatory", False)

            service_path = self.root + self.initdDir + "/" + name
            runlevel_path = self.root + self.runlevelsDir + "/" + runlevel

            if exists(service_path):
                if exists(runlevel_path):
                    ec = target_env_call(["rc-update", state, name, runlevel])
                    if ec != 0:
                        warning("Cannot {} service {} to {}".format(state, name, runlevel))
                        warning("rc-update returned error code {!s}".format(ec))
                        if mandatory:
                            title = _("Cannot modify service")
                            diagnostic = _("<code>rc-update {arg!s}</code> call in chroot returned error code {num!s}.").format(arg=state, num=ec)
                            return (title,
                                    self.make_failure_description(state, name, runlevel) + " " + diagnostic
                                    )
                else:
                    warning("Target runlevel {} does not exist for {}.".format(runlevel, name))
                    if mandatory:
                        title = _("Target runlevel does not exist")
                        diagnostic = _("The path for runlevel {level!s} is <code>{path!s}</code>, which does not exist.").format(level=runlevel, path=runlevel_path)

                        return (title,
                                self.make_failure_description(state, name, runlevel) + " " + diagnostic
                                )
            else:
                warning("Target service {} does not exist in {}.".format(name, self.initdDir))
                if mandatory:
                    title = _("Target service does not exist")
                    diagnostic = _("The path for service {name!s} is <code>{path!s}</code>, which does not exist.").format(name=name, path=service_path)
                    return (title,
                            self.make_failure_description(state, name, runlevel) + " " + diagnostic
                            )


    def run(self):
        """Run the controller
        """

        for state in ("add", "del"):
            r = self.update(state)
            if r is not None:
                return r

def run():
    """
    Setup services
    """

    return OpenrcController().run()



#! /usr/bin/env python3
#
# SPDX-FileCopyrightText: 2020 Adriaan de Groot <groot@kde.org>
# SPDX-License-Identifier: BSD-2-Clause
#
usage = """
Validates a Calamares config file -- YAML syntax -- against a schema.

The schema is also written in YAML syntax, but the schema itself
is JSON-schema. This is possible because all JSON is YAML, and most
YAML is JSON. The limited subset of YAML that Calamares uses is
JSON-representable, anyway.

Usage:
    configvalidator.py <schema> <file> ...
    configvalidator.py -m <module>
    configvalidator.py -x

Exits with value 0 on success, otherwise:
    1 on missing dependencies
    2 on invalid command-line arguments
    3 on missing files
    4 if files have invalid syntax
    5 if files fail to validate
Use -x as only command-line argument to check the imports only.

Use -m <module> as shorthand for standard paths in src/modules/<module>/
"""

# The schemata originally lived outside the Calamares repository,
# without documented tooling. By putting them in the repository
# with the example files and explicit tooling, there's a better
# chance of them catching problems and acting as documentation.

dependencies = """
Dependencies for this tool are: py-yaml and py-jsonschema.

    https://pyyaml.org/
    https://github.com/Julian/jsonschema

Simple installation is `pip install pyyaml jsonschema`
"""

ERR_IMPORT, ERR_USAGE, ERR_FILE_NOT_FOUND, ERR_SYNTAX, ERR_INVALID = range(1,6)

### DEPENDENCIES
#
#
try:
    from jsonschema import validate, SchemaError, ValidationError
    from jsonschema import draft7_format_checker
    from yaml import safe_load, YAMLError
except ImportError as e:
    print(e)
    print(dependencies)
    exit(ERR_IMPORT)

from os.path import exists
import sys

### INPUT VALIDATION
#
#
if len(sys.argv) < 3:
    # Special-case: called with -x to just test the imports
    if len(sys.argv) == 2 and sys.argv[1] == "-x":
        exit(0)
    print(usage)
    exit(ERR_USAGE)

if len(sys.argv) == 3 and sys.argv[1] == "-m":
    module = sys.argv[2]
    schema_file_name = f"src/modules/{module}/{module}.schema.yaml"
    config_file_names = [ f"src/modules/{module}/{module}.conf" ]
else:
    schema_file_name = sys.argv[1]
    config_file_names = sys.argv[2:]

if not exists(schema_file_name):
    print(usage)
    print("\nSchema file '{}' does not exist.".format(schema_file_name))
    exit(ERR_FILE_NOT_FOUND)
for f in config_file_names:
    if not exists(f):
        print(usage)
        print("\nYAML file '{}' does not exist.".format(f))
        exit(ERR_FILE_NOT_FOUND)

### FILES SYNTAX CHECK
#
#
with open(schema_file_name, "r") as data:
    try:
        schema = safe_load(data)
    except YAMLError as e:
        print("Schema error: {} {}.".format(e.problem, e.problem_mark))
        print("\nSchema file '{}' is invalid YAML.".format(schema_file_name))
        exit(ERR_SYNTAX)

try:
    validate(instance={}, schema=schema)
# While developing the schemata, get full exceptions from schema failure
except SchemaError as e:
    print(e)
    print("\nSchema file '{}' is invalid JSON-Schema.".format(schema_file_name))
    exit(ERR_INVALID)
except ValidationError:
    # Just means that empty isn't valid, but the Schema itself is
    pass

configs = []
for f in config_file_names:
    config = None
    with open(f, "r") as data:
        try:
            config = safe_load(data)
        except YAMLError as e:
            print("YAML error: {} {}.".format(e.problem, e.problem_mark))
            print("\nYAML file '{}' is invalid.".format(f))
            exit(ERR_SYNTAX)
    if config is None:
        print("YAML file '{}' is empty.".format(f))
    configs.append(config)

assert len(configs) == len(config_file_names), "Not all configurations loaded."

### SCHEMA VALIDATION
#
#
for c, f in zip(configs, config_file_names):
    try:
        validate(instance=c, schema=schema)
    except ValidationError as e:
        print(e)
        print("\nConfig file '{}' does not validate in schema.".format(f))
        exit(ERR_INVALID)
